import asyncio
import queue
import time

import numpy as np
import pygame
import pygame.locals

from . import rfb

KEYMAP = {
    'bsp': rfb.KEY_BackSpace,
    'tab': rfb.KEY_Tab,
    'return': rfb.KEY_Return,
    'enter': rfb.KEY_Return,
    'esc': rfb.KEY_Escape,
    'ins': rfb.KEY_Insert,
    'delete': rfb.KEY_Delete,
    'del': rfb.KEY_Delete,
    'home': rfb.KEY_Home,
    'end': rfb.KEY_End,
    'pgup': rfb.KEY_PageUp,
    'pgdn': rfb.KEY_PageDown,
    'left': rfb.KEY_Left,
    'up': rfb.KEY_Up,
    'right': rfb.KEY_Right,
    'down': rfb.KEY_Down,

    'slash': rfb.KEY_BackSlash,
    'bslash': rfb.KEY_BackSlash,
    'fslash': rfb.KEY_ForwardSlash,
    'spacebar': rfb.KEY_SpaceBar,
    'space': rfb.KEY_SpaceBar,
    'sb': rfb.KEY_SpaceBar,

    'f1': rfb.KEY_F1,
    'f2': rfb.KEY_F2,
    'f3': rfb.KEY_F3,
    'f4': rfb.KEY_F4,
    'f5': rfb.KEY_F5,
    'f6': rfb.KEY_F6,
    'f7': rfb.KEY_F7,
    'f8': rfb.KEY_F8,
    'f9': rfb.KEY_F9,
    'f10': rfb.KEY_F10,
    'f11': rfb.KEY_F11,
    'f12': rfb.KEY_F12,
    'f13': rfb.KEY_F13,
    'f14': rfb.KEY_F14,
    'f15': rfb.KEY_F15,
    'f16': rfb.KEY_F16,
    'f17': rfb.KEY_F17,
    'f18': rfb.KEY_F18,
    'f19': rfb.KEY_F19,
    'f20': rfb.KEY_F20,

    'lshift': rfb.KEY_ShiftLeft,
    'shift': rfb.KEY_ShiftLeft,
    'rshift': rfb.KEY_ShiftRight,
    'lctrl': rfb.KEY_ControlLeft,
    'ctrl': rfb.KEY_ControlLeft,
    'rctrl': rfb.KEY_ControlRight,
    'lmeta': rfb.KEY_MetaLeft,
    'meta': rfb.KEY_MetaLeft,
    'rmeta': rfb.KEY_MetaRight,
    'lalt': rfb.KEY_AltLeft,
    'alt': rfb.KEY_AltLeft,
    'ralt': rfb.KEY_AltRight,
    'scrlk': rfb.KEY_Scroll_Lock,
    'sysrq': rfb.KEY_Sys_Req,
    'numlk': rfb.KEY_Num_Lock,
    'caplk': rfb.KEY_Caps_Lock,
    'pause': rfb.KEY_Pause,
    'lsuper': rfb.KEY_Super_L,
    'super': rfb.KEY_Super_L,
    'rsuper': rfb.KEY_Super_R,
    'lhyper': rfb.KEY_Hyper_L,
    'hyper': rfb.KEY_Hyper_L,
    'rhyper': rfb.KEY_Hyper_R,

    'kp0': rfb.KEY_KP_0,
    'kp1': rfb.KEY_KP_1,
    'kp2': rfb.KEY_KP_2,
    'kp3': rfb.KEY_KP_3,
    'kp4': rfb.KEY_KP_4,
    'kp5': rfb.KEY_KP_5,
    'kp6': rfb.KEY_KP_6,
    'kp7': rfb.KEY_KP_7,
    'kp8': rfb.KEY_KP_8,
    'kp9': rfb.KEY_KP_9,
    'kpenter': rfb.KEY_KP_Enter,
}

PYGAME_KEYMAP = {
    pygame.K_BACKSPACE: 'bsp',
    pygame.K_TAB: 'tab',
    pygame.K_RETURN: 'return',
    pygame.K_ESCAPE: 'esc',
    pygame.K_INSERT: 'ins',
    pygame.K_DELETE: 'delete',
    pygame.K_HOME: 'home',
    pygame.K_END: 'end',
    pygame.K_PAGEUP: 'pgup',
    pygame.K_PAGEDOWN: 'pgdn',

    pygame.K_LEFT: 'left',
    pygame.K_UP: 'up',
    pygame.K_RIGHT: 'right',
    pygame.K_DOWN: 'down',
    pygame.K_SLASH: 'slash',
    pygame.K_BACKSLASH: 'bslash',
    pygame.K_SPACE: 'space',

    pygame.K_F1: 'f1',
    pygame.K_F2: 'f2',
    pygame.K_F3: 'f3',
    pygame.K_F4: 'f4',
    pygame.K_F5: 'f5',
    pygame.K_F6: 'f6',
    pygame.K_F7: 'f7',
    pygame.K_F8: 'f8',
    pygame.K_F9: 'f9',
    pygame.K_F10: 'f10',
    pygame.K_F11: 'f11',
    pygame.K_F12: 'f12',
    pygame.K_F13: 'f13',
    pygame.K_F14: 'f14',
    pygame.K_F15: 'f15',

    pygame.K_LSHIFT: 'lshift',
    pygame.K_RSHIFT: 'rshift',
    pygame.K_LCTRL: 'lctrl',
    pygame.K_RCTRL: 'rctrl',
    pygame.K_LMETA: 'lmeta',
    pygame.K_RMETA: 'rmeta',
    pygame.K_LALT: 'lalt',
    pygame.K_RALT: 'ralt',
    pygame.K_SCROLLOCK: 'scrlk',
    pygame.K_SYSREQ: 'sysrq',
    pygame.K_NUMLOCK: 'numlk',
    pygame.K_CAPSLOCK: 'caplk',
    pygame.K_PAUSE: 'pause',
    pygame.K_LSUPER: 'lsuper',
    pygame.K_RSUPER: 'rsuper',
    pygame.K_KP0: 'kp0',
    pygame.K_KP1: 'kp1',
    pygame.K_KP2: 'kp2',
    pygame.K_KP3: 'kp3',
    pygame.K_KP4: 'kp4',
    pygame.K_KP5: 'kp5',
    pygame.K_KP6: 'kp6',
    pygame.K_KP7: 'kp7',
    pygame.K_KP8: 'kp8',
    pygame.K_KP9: 'kp9',
    pygame.K_KP_ENTER: 'kpenter',
}


# KEYMAP = {k: KEYMAP.get(v) for k, v in PYGAME_KEYMAP.items()}


class AuthenticationError(Exception):
    ...


class VNCClient(rfb.RFBClient):
    # RAW_ENCODING = 0
    # COPY_RECTANGLE_ENCODING = 1
    # RRE_ENCODING = 2
    # CORRE_ENCODING = 4
    # HEXTILE_ENCODING = 5
    # ZLIB_ENCODING = 6
    # TIGHT_ENCODING = 7
    # ZLIBHEX_ENCODING = 8
    # ZRLE_ENCODING = 16
    encoding = rfb.ZRLE_ENCODING
    x = 0
    y = 0
    buttons = 0
    screen = None
    image_mode = "RGBX"
    deferred = None

    cursor = None
    cmask = None

    SPECIAL_KEYS_US = "~!@#$%^&*()_+{}|:\"<>?"

    def __init__(self, host, port, loop):
        super(VNCClient, self).__init__(host, port, loop)

        self.force_caps = False
        self.nocursor = False
        self.pseudodesktop = False
        self.pseudocursor = False
        self.shared = True

        self.gui_events = queue.Queue()
        self.mouse_events = queue.Queue()
        self.keyboard_events = queue.Queue()

    def _decode_key(self, key):
        if self.force_caps:
            if key.isupper() or key in self.SPECIAL_KEYS_US:
                key = 'shift-%c' % key

        if len(key) == 1:
            keys = [key]
        else:
            keys = key.split('-')

        keys = [KEYMAP.get(k) or ord(k) for k in keys]

        return keys

    async def _gui_handler(self):
        pygame.init()

        surface = pygame.display.set_mode((self.width, self.height))
        pygame.mouse.set_visible(0)

        while True:
            await asyncio.sleep(0.01, loop=self.loop)

            gui_event = None
            while not self.gui_events.empty():
                gui_event = self.gui_events.get_nowait()
            if gui_event is not None:
                screen, = gui_event
                import cv2
                screen = cv2.cvtColor(screen, cv2.COLOR_BGRA2RGB)
                surf = pygame.surfarray.make_surface(np.swapaxes(np.asarray(screen), 0, 1))
                surface.blit(surf, (0, 0))
            pygame.display.update()
            pygame.display.flip()

            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    exit(1)
                elif event.type == pygame.KEYDOWN:
                    mod = event.mod
                    key = event.key

                    self.keyboard_events.put_nowait(('down', mod, key))
                elif event.type == pygame.KEYUP:
                    mod = event.mod
                    key = event.key

                    self.keyboard_events.put_nowait(('up', mod, key))
                elif event.type == pygame.MOUSEMOTION:
                    x, y = pygame.mouse.get_pos()

                    self.mouse_events.put_nowait(('move', x, y))
                elif event.type == pygame.MOUSEBUTTONDOWN:
                    button = event.button

                    self.mouse_events.put_nowait(('down', button))
                elif event.type == pygame.MOUSEBUTTONUP:
                    button = event.button

                    self.mouse_events.put_nowait(('up', button))

    async def _mouse_handler(self):
        while True:
            await asyncio.sleep(0.01, loop=self.loop)

            if self.mouse_events.empty():
                continue

            mouse_event = self.mouse_events.get_nowait()
            if mouse_event[0] == 'move':
                x, y = mouse_event[1:]
                await self.mouse_move(x, y)
            if mouse_event[0] == 'down':
                button, = mouse_event[1:]
                await self.mouse_down(button)
            if mouse_event[0] == 'up':
                button, = mouse_event[1:]
                await self.mouse_up(button)

    async def _keyboard_handler(self):
        while True:
            await asyncio.sleep(0.01, loop=self.loop)

            if self.keyboard_events.empty():
                continue

            keyboard_event = self.keyboard_events.get_nowait()
            mod, key = keyboard_event[1:]
            if ord('a') <= key <= ord('z'):
                key = chr(key)
            elif ord('A') <= key <= ord('Z'):
                key = chr(key)
            elif key in PYGAME_KEYMAP.keys():
                key = PYGAME_KEYMAP.get(key)
            else:
                continue

            if keyboard_event[0] == 'down':
                await self.key_down(key)
            if keyboard_event[0] == 'up':
                await self.key_up(key)

    async def pause(self, duration):
        await asyncio.sleep(duration, loop=self.loop)

    async def key_press(self, key):
        print('keyPress %s', key)
        await self.key_down(key)
        await self.key_up(key)

    async def key_down(self, key):
        print('keyDown %s', key)
        keys = self._decode_key(key)
        for k in keys:
            await self.key_event(k, down=1)

    async def key_up(self, key):
        print('keyUp %s', key)
        keys = self._decode_key(key)
        for k in keys:
            await self.key_event(k, down=0)

    async def mouse_press(self, button):
        print('mousePress %s', button)
        buttons = self.buttons | (1 << (button - 1))
        await self.mouse_down(button)
        await self.mouse_up(button)

    async def mouse_down(self, button):
        print('mouseDown %s', button)
        self.buttons |= 1 << (button - 1)
        await self.pointer_event(self.x, self.y, buttonmask=self.buttons)

    async def mouse_up(self, button):
        print('mouseUp %s', button)
        self.buttons &= ~(1 << (button - 1))
        await self.pointer_event(self.x, self.y, buttonmask=self.buttons)

    async def capture_screen(self, filename, incremental=0):
        print('captureScreen %s', filename)
        return await self._capture(filename, incremental)

    async def capture_region(self, filename, x, y, w, h, incremental=0):
        print('captureRegion %s', filename)
        return self._capture(filename, incremental, x, y, x + w, y + h)

    async def refresh_screen(self, incremental=0):
        await self.framebuffer_update_request(incremental=incremental)

    async def _capture(self, filename, incremental, *args):
        await self.refresh_screen(incremental)
        await self._capture_save(filename)

    async def _capture_save(self, filename, *args):
        # print('captureSave %s', filename)
        # if args:
        #     capture = self.screen.crop(args)
        # else:
        #     capture = self.screen
        # capture.save(filename)
        ...

    async def expect_screen(self, filename, maxrms=0):
        print('expectScreen %s', filename)
        return self._expect_framebuffer(filename, 0, 0, maxrms)

    async def expect_region(self, filename, x, y, maxrms=0):
        print('expectRegion %s (%s, %s)', filename, x, y)
        return self._expect_framebuffer(filename, x, y, maxrms)

    async def _expect_framebuffer(self, filename, x, y, maxrms):
        # await self.framebuffer_update_request(incremental=1)
        # image = Image.open(filename)
        # w, h = image.size
        # self.expected = image.histogram()
        # await self._expect_compare((x, y, x + w, y + h), maxrms)
        ...

    async def _expect_compare(self, box, maxrms):
        # image = self.screen.crop(box)
        #
        # hist = image.histogram()
        # if len(hist) == len(self.expected):
        #     sum_ = 0
        #     for h, e in zip(hist, self.expected):
        #         sum_ += (h - e) ** 2
        #     rms = math.sqrt(sum_ / len(hist))
        #
        #     print('rms:%s maxrms: %s', int(rms), int(maxrms))
        #     if rms <= maxrms:
        #         return self
        #
        # await self._expect_compare(box, maxrms)
        # await self.framebuffer_update_request(incremental=1)
        ...

    async def mouse_move(self, x, y):
        print('mouseMove %d,%d', x, y)
        self.x, self.y = x, y
        await self.pointer_event(x, y, self.buttons)

    async def mouse_drag(self, x, y, step=1):
        print('mouseDrag %d,%d', x, y)
        if x < self.x:
            xsteps = [self.x - i for i in range(step, self.x - x + 1, step)]
        else:
            xsteps = range(self.x, x, step)

        if y < self.y:
            ysteps = [self.y - i for i in range(step, self.y - y + 1, step)]
        else:
            ysteps = range(self.y, y, step)

        for ypos in ysteps:
            await self.mouse_move(self.x, ypos)
            time.sleep(.2)

        for xpos in xsteps:
            await self.mouse_move(xpos, self.y)
            time.sleep(.2)

        await self.mouse_move(x, y)

    async def set_image_mode(self):
        if self.server_version == 3.889:
            await self.set_pixel_format(bpp=16, depth=16, bigendian=0, truecolor=1, redmax=31, greenmax=63, bluemax=31, redshift=11, greenshift=5, blueshift=0)
            self.image_mode = "BGR;16"
        elif self.truecolor and (not self.bigendian) and self.depth == 24 and self.redmax == 255 and self.greenmax == 255 and self.bluemax == 255:
            pixel = ["X"] * self.bypp
            offsets = [offset // 8 for offset in (self.redshift, self.greenshift, self.blueshift)]
            for offset, color in zip(offsets, "RGB"):
                pixel[offset] = color
            self.image_mode = "".join(pixel)
        else:
            await self.set_pixel_format()

    async def vnc_connection_made(self):
        await self.set_image_mode()
        encodings = [self.encoding]
        if self.pseudocursor or self.nocursor:
            encodings.append(rfb.PSEUDO_CURSOR_ENCODING)
        if self.pseudodesktop:
            encodings.append(rfb.PSEUDO_DESKTOP_SIZE_ENCODING)
        await self.set_encodings(encodings)

        asyncio.ensure_future(self._gui_handler(), loop=self.loop)
        asyncio.ensure_future(self._mouse_handler(), loop=self.loop)
        asyncio.ensure_future(self._keyboard_handler(), loop=self.loop)

    async def vnc_request_password(self):
        if self.password is None:
            await self.writer.close()
            print('password required, but none provided')
            return
        await self.send_password(self.password)

    async def vnc_auth_failed(self, reason):
        print("Cannot connect %s" % reason)

    async def begin_update(self):
        ...

    async def commit_update(self, rectangles=None):
        ...

    async def update_rectangle(self, x, y, width, height, data):
        if not data:
            return

        update = np.frombuffer(data, dtype='uint8').reshape((height, width, 4))
        if self.screen is None:
            self.screen = np.zeros((self.height, self.width, 4), dtype='uint8')
        else:
            self.screen[y: y + update.shape[0], x: x + update.shape[1]] = update

        self.gui_events.put_nowait((self.screen,))

        if not self.rectangles:
            import datetime
            print(datetime.datetime.utcnow().isoformat())
            await self.refresh_screen()

    async def copy_rectangle(self, srcx, srcy, x, y, width, height):
        ...

    async def fill_rectangle(self, x, y, width, height, color):
        await self.update_rectangle(x, y, width, height, color * width * height)

    async def update_cursor(self, x, y, width, height, image, mask):
        if self.nocursor:
            return

        if not width or not height:
            self.cursor = None

        # self.cursor = Image.frombytes('RGBX', (width, height), image)
        # self.cmask = Image.frombytes('1', (width, height), mask)
        self.cfocus = x, y
        await self.draw_cursor()

    async def draw_cursor(self):
        if not self.cursor:
            return

        if not self.screen:
            return

        # x = self.x - self.cfocus[0]
        # y = self.y - self.cfocus[1]
        # self.screen.paste(self.cursor, (x, y), self.cmask)

    async def update_desktop_size(self, width, height):
        # new_screen = Image.new("RGB", (width, height), "black")
        # if self.screen:
        #     new_screen.paste(self.screen, (0, 0))
        # self.screen = new_screen
        ...

    async def bell(self):
        print('ding')

    async def copy_text(self, text):
        print('clipboard copy', repr(text))

    async def paste(self, message):
        await self.client_cut_text(message)



from joatmon.nn import functional as f
from joatmon.nn.core import Optimizer

__all__ = ['RMSprop']


class RMSprop(Optimizer):
    def __init__(self, params, lr=1e-2, alpha=0.99, eps=1e-8, weight_decay=0, momentum=0, centered=False):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= momentum:
            raise ValueError("Invalid momentum value: {}".format(momentum))
        if not 0.0 <= weight_decay:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))
        if not 0.0 <= alpha:
            raise ValueError("Invalid alpha value: {}".format(alpha))

        defaults = dict(lr=lr, momentum=momentum, alpha=alpha, eps=eps, centered=centered, weight_decay=weight_decay)
        super(RMSprop, self).__init__(params, defaults)

    def __setstate__(self, state):
        super(RMSprop, self).__setstate__(state)
        for group in self.param_groups:
            group.setdefault('momentum', 0)
            group.setdefault('centered', False)

    def step(self):
        for group in self.param_groups:
            params_with_grad = []
            grads = []
            square_avgs = []
            alphas = []
            momentum_buffers = []
            grad_avgs = []

            for p in group['params']:
                if p.grad is not None:
                    params_with_grad.append(p)
                    grads.append(p.grad)

                    state = self.state[p]
                    if len(state) == 0:
                        state['step'] = 0
                        state['square_avg'] = f.zeros_like(p)
                        if group['momentum'] > 0:
                            state['momentum_buffer'] = f.zeros_like(p)
                        if group['centered']:
                            state['grad_avg'] = f.zeros_like(p)

                    square_avgs.append(state['square_avg'])
                    alphas.append(group['alpha'])

                    if group['momentum'] > 0:
                        momentum_buffers.append(state['momentum_buffer'])
                    if group['centered'] > 0:
                        grad_avgs.append(state['grad_avg'])

                    state['step'] += 1
            f.rmsprop(
                params_with_grad, grads, square_avgs, alphas, momentum_buffers,
                grad_avgs, group['momentum'], group['centered'], group['lr'], group['weight_decay'], group['eps']
            )



import torch
import torch.nn as nn
import torch.nn.functional as F


class PPOActor(nn.Module):
    def __init__(self, in_features, out_features):
        super(PPOActor, self).__init__()

        self.hidden1 = nn.Linear(in_features, 128)
        self.hidden2 = nn.Linear(128, 128)
        self.out = nn.Linear(128, out_features)

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        x = F.relu(self.hidden1(state))
        x = F.relu(self.hidden2(x))
        action = self.out(x).tanh()

        return action


class PPOCritic(nn.Module):
    def __init__(self, in_features, out_features):
        super(PPOCritic, self).__init__()

        self.hidden1 = nn.Linear(in_features + out_features, 128)
        self.hidden2 = nn.Linear(128, 128)
        self.out = nn.Linear(128, 1)

    def forward(self, state: torch.Tensor, action: torch.Tensor) -> torch.Tensor:
        x = torch.cat((state, action), dim=-1)
        x = F.relu(self.hidden1(x))
        x = F.relu(self.hidden2(x))
        value = self.out(x)

        return value



import json
import os

from elevenlabs import (
    generate,
    set_api_key
)


class TTSAgent:
    def __init__(self):
        super(TTSAgent, self).__init__()

        set_api_key(json.loads(open('iva/iva.json', 'r').read())['config']['elevenlabs']['key'])

    def convert(self, text):
        text_hash = ''.join([c.lower() for c in text.lower().strip() if c not in ['!', '?', ',', '/', '\\']]).replace(' ', '_').replace('.', '_').replace(':', '_')
        text_audio_path = os.path.join('iva/cache', f'{text_hash}.mp3')

        if not os.path.exists(os.path.dirname(text_audio_path)):
            os.makedirs(os.path.dirname(text_audio_path))

        if os.path.exists(text_audio_path):
            audio = open(text_audio_path, 'rb').read()
        else:
            audio = generate(
                text=text,
                voice=json.loads(open('iva/iva.json', 'r').read())['config']['elevenlabs']['voice'],
                model=json.loads(open('iva/iva.json', 'r').read())['config']['elevenlabs']['model']
            )

        return audio



import marshal
import random
import xml.etree.ElementTree as ElementTree

import cv2
import numpy as np
import pygame
import pymunk
from pygame.color import THECOLORS
from pymunk.body import Body
from pymunk.vec2d import Vec2d

__all__ = ['SokobanEnv']

from joatmon.game.core import CoreEnv

TYPE_LOOKUP = {
    0: 'wall',
    1: 'empty space',
    2: 'box target',
    3: 'box on target',
    4: 'box not on target',
    5: 'player'
}

ACTION_LOOKUP = {
    0: 'push up',
    1: 'push down',
    2: 'push left',
    3: 'push right',
    4: 'move up',
    5: 'move down',
    6: 'move left',
    7: 'move right',
}

CHANGE_COORDINATES = {
    0: (-1, 0),
    1: (1, 0),
    2: (0, -1),
    3: (0, 1)
}


def generate_room(dim=(7, 7), wall_prob=0.3, p_change_directions=0.35, num_steps=5, num_boxes=1, tries=4):
    def room_topology_generation(_dim=(10, 10), _wall_prob=0.1, _p_change_directions=0.35, _num_steps=15):
        dim_x, dim_y = _dim

        masks = [
            [[0, 0, 0], [1, 1, 1], [0, 0, 0]],
            [[0, 1, 0], [0, 1, 0], [0, 1, 0]],
            [[0, 0, 0], [1, 1, 0], [0, 1, 0]],
            [[0, 0, 0], [1, 1, 0], [1, 1, 0]],
            [[0, 0, 0], [0, 1, 1], [0, 1, 0]]
        ]

        directions = [(1, 0), (0, 1), (-1, 0), (0, -1)]
        direction = random.sample(directions, 1)[0]

        position = np.array([random.randint(1, dim_x - 1), random.randint(1, dim_y - 1)])

        level = np.random.choice(a=[0, 1], size=_dim, p=[_wall_prob, 1.0 - _wall_prob]).astype('int')
        # for y in range(dim_y):
        #     for x in range(dim_x):
        #         if random.random() < wall_prob:
        #             level[y][x] = 0

        for s in range(_num_steps):
            if random.random() < _p_change_directions:
                direction = random.sample(directions, 1)[0]

            position = position + direction
            position[0] = max(min(position[0], dim_x - 2), 1)
            position[1] = max(min(position[1], dim_y - 2), 1)

            mask = random.sample(masks, 1)[0]
            mask_start = position - 1
            level[mask_start[0]:mask_start[0] + 3, mask_start[1]:mask_start[1] + 3] += mask

        level[level > 0] = 1
        level[:, [0, dim_y - 1]] = 0
        level[[0, dim_x - 1], :] = 0

        return level

    def place_boxes_and_player(_room, _num_boxes):
        possible_positions = np.where(_room == 1)
        num_possible_positions = possible_positions[0].shape[0]
        num_players = 1

        if num_possible_positions <= _num_boxes + num_players:
            raise RuntimeError('Not enough free spots (#{}) to place {} player and {} boxes.'.format(num_possible_positions, num_players, _num_boxes))

        ind = np.random.randint(num_possible_positions)
        player_position = possible_positions[0][ind], possible_positions[1][ind]
        _room[player_position] = 5

        for n in range(_num_boxes):
            possible_positions = np.where(_room == 1)
            num_possible_positions = possible_positions[0].shape[0]

            ind = np.random.randint(num_possible_positions)
            box_position = possible_positions[0][ind], possible_positions[1][ind]
            _room[box_position] = 2

        return _room

    global_explored_states = set()
    global_num_boxes = 0
    global_best_room_score = -1
    global_best_room = None
    global_best_box_mapping = None

    def reverse_playing(_room_state, _room_structure):
        global global_explored_states, global_num_boxes, global_best_room_score, global_best_room, global_best_box_mapping

        box_mapping = {}
        box_locations = np.where(_room_structure == 2)
        global_num_boxes = len(box_locations[0])
        for l in range(global_num_boxes):
            box = (box_locations[0][l], box_locations[1][l])
            box_mapping[box] = box

        global_explored_states = set()
        global_best_room_score = -1
        global_best_box_mapping = box_mapping
        depth_first_search(_room_state, _room_structure, box_mapping, last_pull=(-1, -1))

        return global_best_room, global_best_room_score, global_best_box_mapping

    def depth_first_search(_room_state, _room_structure, _box_mapping, box_swaps=0, last_pull=(-1, -1), ttl=300):
        global global_explored_states, global_num_boxes, global_best_room_score, global_best_room, global_best_box_mapping

        ttl -= 1
        if ttl <= 0 or len(global_explored_states) >= 300000:
            return

        state_to_hash = marshal.dumps(_room_state)

        if not (state_to_hash in global_explored_states):
            room_score = box_swaps * box_displacement_score(_box_mapping)
            if np.where(_room_state == 2)[0].shape[0] != global_num_boxes:
                room_score = 0

            if room_score > global_best_room_score:
                global_best_room = _room_state
                global_best_room_score = room_score
                global_best_box_mapping = _box_mapping

            global_explored_states.add(state_to_hash)

            for action in ACTION_LOOKUP.keys():
                room_state_next = _room_state.copy()
                box_mapping_next = _box_mapping.copy()

                room_state_next, box_mapping_next, last_pull_next = reverse_move(room_state_next, _room_structure, box_mapping_next, last_pull, action)

                box_swaps_next = box_swaps
                if last_pull_next != last_pull:
                    box_swaps_next += 1

                depth_first_search(room_state_next, _room_structure, box_mapping_next, box_swaps_next, last_pull, ttl)

    def reverse_move(_room_state, _room_structure, _box_mapping, last_pull, action):
        player_position = np.where(_room_state == 5)
        player_position = np.array([player_position[0][0], player_position[1][0]])

        change = CHANGE_COORDINATES[action % 4]
        next_position = player_position + change

        if _room_state[next_position[0], next_position[1]] in [1, 2]:
            _room_state[player_position[0], player_position[1]] = _room_structure[player_position[0], player_position[1]]
            _room_state[next_position[0], next_position[1]] = 5

            if action < 4:
                possible_box_location = change[0] * -1, change[1] * -1
                possible_box_location += player_position

                if _room_state[possible_box_location[0], possible_box_location[1]] in [3, 4]:
                    _room_state[player_position[0], player_position[1]] = 3
                    _room_state[possible_box_location[0], possible_box_location[1]] = _room_structure[possible_box_location[0], possible_box_location[1]]

                    for k in _box_mapping.keys():
                        if _box_mapping[k] == (possible_box_location[0], possible_box_location[1]):
                            _box_mapping[k] = (player_position[0], player_position[1])
                            last_pull = k

        return _room_state, _box_mapping, last_pull

    def box_displacement_score(_box_mapping):
        score = 0

        for box_target in _box_mapping.keys():
            box_location = np.array(_box_mapping[box_target])
            box_target = np.array(box_target)
            dist = np.sum(np.abs(box_location - box_target))
            score += dist

        return score

    room_state = np.zeros(shape=dim)
    room_structure = np.zeros(shape=dim)

    score = 0
    box_mapping = None
    for t in range(tries):
        room = room_topology_generation(dim, wall_prob, p_change_directions, num_steps)
        room = place_boxes_and_player(room, _num_boxes=num_boxes)

        room_structure = np.copy(room)
        room_structure[room_structure == 5] = 1

        room_state = room.copy()
        room_state[room_state == 2] = 4

        room_state, score, box_mapping = reverse_playing(room_state, room_structure)
        room_state[room_state == 3] = 4

        if score > 0:
            break

    if score == 0:
        raise RuntimeWarning('Generated Model with score == 0')

    return room_structure, room_state, box_mapping


def create_circle_body(mass, body_type, radius):
    moment = pymunk.moment_for_circle(mass, 0.0, radius)
    body = pymunk.Body(mass, moment, body_type)
    return body


def create_circle_shape(body, radius, friction, elasticity, collision_type, sensor):
    shape = pymunk.Circle(body, radius)
    shape.friction = friction
    shape.elasticity = elasticity
    shape.collision_type = collision_type
    shape.sensor = sensor
    return shape


def create_rectangle_body(mass, body_type, half_size):
    points = [(-half_size, -half_size), (-half_size, half_size), (half_size, half_size), (half_size, -half_size)]
    moment = pymunk.moment_for_poly(mass, points)
    body = pymunk.Body(mass, moment, body_type)
    return body, points


def create_rectangle_shape(body, points, friction, elasticity, collision_type, sensor):
    shape = pymunk.Poly(body, points)
    shape.friction = friction
    shape.elasticity = elasticity
    shape.collision_type = collision_type
    shape.sensor = sensor
    return shape


def draw_circle(screen, color, position, radius):
    position = flip_y(position, screen.get_size()[1])
    pygame.draw.circle(screen, color, position.int_tuple, int(radius))


def draw_rectangle(screen, color, position, half_size):
    position = flip_y(position, screen.get_size()[1])
    points = [
        (-half_size + position.x, -half_size + position.y),
        (-half_size + position.x, half_size + position.y),
        (half_size + position.x, half_size + position.y),
        (half_size + position.x, -half_size + position.y)
    ]
    pygame.draw.polygon(screen, color, points)


def draw_sprite(screen, image, position, half_size):
    screen.blit(image, flip_y(position, screen.get_size()[1]) - Vec2d(half_size, half_size))


def flip_y(vector, y):
    return Vec2d(int(vector.x), int(-vector.y + y))


def layout_getter(layout_specs):
    def _get_layout():
        return generate_room(
            dim=(int(layout_specs['height']), int(layout_specs['width'])),
            wall_prob=float(layout_specs['wall_prob']),
            p_change_directions=float(layout_specs['p_change_directions']),
            num_steps=int(layout_specs['num_steps']),
            num_boxes=int(layout_specs['num_boxes']),
            tries=int(layout_specs['tries'])
        )[1]

    return _get_layout


def load_sprite(path, color, size=None):
    # if not existing use color, get color as parameter
    try:
        sprite = pygame.image.load(path)
        if size is not None:
            if isinstance(size, (list, tuple)):
                if len(size) != 2:
                    raise Exception
            elif isinstance(size, (float, int)):
                size = (int(size), int(size))
        sprite = pygame.transform.scale(sprite, size)
    except Exception as ex:
        print(str(ex))
        sprite = None

    if sprite is None:
        sprite = pygame.Surface((size, size))
        sprite.fill(color)

    return sprite


def load_xml(element):
    # can be common utility
    try:
        if isinstance(element, str):
            return load_xml(ElementTree.parse(element).getroot())
    except Exception as ex:
        print(str(ex))
        return {}

    d = {}
    for child in element:
        if len(child):
            d['{}'.format(child.tag)] = load_xml(child)
        else:
            d['{}'.format(child.tag)] = child.text

    return d


def euclidean_distance(position1, position2):
    return ((position1.x - position2.x) ** 2 + (position1.y - position2.y) ** 2) ** 0.5


def manhattan_distance(position1, position2):
    return abs(position1.x - position2.x) + abs(position1.y - position2.y)


def convert_to(value, type_):
    try:
        if isinstance(value, str):
            return type_(value)
        else:
            return value
    except ValueError:
        raise ValueError


class SokobanEnv(CoreEnv):
    def __init__(self, xml, xmls, sprites):
        super().__init__()

        env_specs = load_xml(xmls + xml)
        if env_specs is None:
            raise Exception

        layout_specs = env_specs.get(
            'layout', {
                'height': 7,
                'width': 7,
                'wall_prob': 0.3,
                'p_change_directions': 0.35,
                'num_steps': 5,
                'num_boxes': 1,
                'tries': 4
            }
        )
        object_specs = env_specs.get(
            'object', {
                'size': 32,
                'ground': {
                    'collision': 0,
                    'color': 'black',
                    'sprite': 'ground.png'
                },
                'goal': {
                    'collision': 1,
                    'color': 'green3',
                    'sprite': 'goal.png'
                },
                'block': {
                    'collision': 2,
                    'color': 'blue3',
                    'sprite': 'block.png'
                },
                'player': {
                    'collision': 4,
                    'color': 'orange3',
                    'sprite': 'player.png'
                },
                'obstacle': {
                    'collision': 8,
                    'color': 'red3',
                    'sprite': 'obstacle.png'
                }
            }
        )
        ground_specs = object_specs.get(
            'ground', {
                'collision': 0,
                'color': 'black',
                'sprite': 'ground.png'
            }
        )
        goal_specs = object_specs.get(
            'goal', {
                'collision': 1,
                'color': 'green3',
                'sprite': 'goal.png'
            }
        )
        block_specs = object_specs.get(
            'block', {
                'collision': 2,
                'color': 'blue3',
                'sprite': 'block.png'
            }
        )
        player_specs = object_specs.get(
            'player', {
                'collision': 4,
                'color': 'orange3',
                'sprite': 'player.png'
            }
        )
        obstacle_specs = object_specs.get(
            'obstacle', {
                'collision': 8,
                'color': 'red3',
                'sprite': 'obstacle.png'
            }
        )
        space_specs = env_specs.get(
            'space', {
                'gravity': {
                    'x': 0.0,
                    'y': 0.0
                },
                'damping': 0.0,
                'dt': 0.02,
                'steps': 2,
                'force_multiplier': 250.0
            }
        )
        gravity_specs = space_specs.get(
            'gravity', {
                'x': 0.0,
                'y': 0.0
            }
        )
        reward_specs = env_specs.get(
            'reward', {
                'movement': -0.25,
                'on_target': 5.0,
                'off_target': -5.0,
                'success': 10.0
            }
        )

        self.new_layout = layout_getter(layout_specs)
        self.layout = None

        self.obj_size = convert_to(object_specs.get('size', 32), int)
        self.world_metrics = (convert_to(layout_specs.get('width', 7), int) * self.obj_size, convert_to(layout_specs.get('height', 7), int) * self.obj_size)

        self.ground_color = THECOLORS[str(ground_specs.get('color'))]

        self.goal_collision = convert_to(goal_specs.get('collision', 1), int)
        self.goal_path = sprites + goal_specs.get('sprite')
        self.goal_color = THECOLORS[str(goal_specs.get('color'))]
        self.goal_sprite = load_sprite(path=self.goal_path, color=self.goal_color, size=self.obj_size)

        self.block_collision = convert_to(block_specs.get('collision', 2), int)
        self.block_path = sprites + block_specs.get('sprite')
        self.block_color = THECOLORS[str(block_specs.get('color'))]
        self.block_sprite = load_sprite(path=self.block_path, color=self.block_color, size=self.obj_size)

        self.player_collision = convert_to(player_specs.get('collision', 4), int)
        self.player_path = sprites + player_specs.get('sprite')
        self.player_color = THECOLORS[str(player_specs.get('color'))]
        self.player_sprite = load_sprite(path=self.player_path, color=self.player_color, size=self.obj_size)

        self.obstacle_collision = convert_to(obstacle_specs.get('collision', 8), int)
        self.obstacle_path = sprites + obstacle_specs.get('sprite')
        self.obstacle_color = THECOLORS[str(obstacle_specs.get('color'))]
        self.obstacle_sprite = load_sprite(path=self.obstacle_path, color=self.obstacle_color, size=self.obj_size)

        self.gravity = (convert_to(gravity_specs.get('x', 0.0), float), convert_to(gravity_specs.get('y', 0.0), float))
        self.damping = convert_to(space_specs.get('damping', 0.0), float)
        self.dt = convert_to(space_specs.get('dt', 0.02), float)
        self.steps = convert_to(space_specs.get('steps', 2), int)
        self.force = self.obj_size * convert_to(space_specs.get('force_multiplier', 250.0), float)

        self.movement_penalty = convert_to(reward_specs.get('movement', -0.25), float)
        self.box_on_target = convert_to(reward_specs.get('on_target', 5.0), float)
        self.box_off_target = convert_to(reward_specs.get('off_target', -5.0), float)
        self.all_boxes_on_target = convert_to(reward_specs.get('success', 10.0), float)

        self._init_space()
        self._init_window()

    def _begin(self, arbiter, _, __):
        if arbiter.shapes[0].collision_type == self.goal_collision:
            goal = arbiter.shapes[0]
            block = arbiter.shapes[1]
        else:
            goal = arbiter.shapes[1]
            block = arbiter.shapes[0]

        if (goal.body, block.body) not in self.overlaps.keys():
            self.overlaps[(goal.body, block.body)] = 0
        return True

    def _clear(self):
        self.obstacles = []
        self.players = []
        self.blocks = []
        self.goals = []

        self.player = None

        self.level_complete = False
        self.overlaps = {}
        self.space.remove(*self.space.shapes, *self.space.bodies)

    def _check_complete(self):
        self.level_complete = self.successes == len(self.goals)

    def _create_objects(self):
        for y, row in enumerate(self.layout.tolist()):
            for x, object_type in enumerate(row):
                object_position = flip_y(vector=Vec2d(self.obj_size * (x + 0.5), self.obj_size * (y + 0.5)), y=self.screen.get_size()[1])

                if object_type == 0:
                    body, points = create_rectangle_body(mass=1000000, body_type=Body.STATIC, half_size=self.obj_size * 0.5)
                    body.position = Vec2d(object_position.x, object_position.y)
                    shape = create_rectangle_shape(body=body, points=points, friction=0, elasticity=0, collision_type=self.obstacle_collision, sensor=False)
                elif object_type == 2:
                    body, points = create_rectangle_body(mass=1000000, body_type=Body.STATIC, half_size=self.obj_size * 0.5)
                    body.position = Vec2d(object_position.x, object_position.y)
                    shape = create_rectangle_shape(body=body, points=points, friction=0, elasticity=0, collision_type=self.goal_collision, sensor=True)
                elif object_type == 4:
                    body, points = create_rectangle_body(mass=1, body_type=Body.DYNAMIC, half_size=self.obj_size * 0.5)
                    body.position = Vec2d(object_position.x, object_position.y)
                    shape = create_rectangle_shape(body=body, points=points, friction=0, elasticity=0, collision_type=self.block_collision, sensor=False)
                elif object_type == 5:
                    body = create_circle_body(mass=1, body_type=Body.DYNAMIC, radius=self.obj_size * 0.5)
                    body.position = Vec2d(object_position.x, object_position.y)
                    shape = create_circle_shape(body=body, radius=self.obj_size * 0.5, friction=0, elasticity=0, collision_type=self.player_collision, sensor=False)
                else:
                    continue

                self.space.add(body)
                self.space.add(shape)

    def _draw(self):
        # clear the screen
        self.screen.fill(self.ground_color)
        self._draw_shapes()

    def _draw_shapes(self):
        for obstacle in self.obstacles:
            if self.obstacle_sprite is not None:
                draw_sprite(self.screen, image=self.obstacle_sprite, position=obstacle.body.position, half_size=self.obj_size * 0.5)
            else:
                draw_rectangle(self.screen, color=self.obstacle_color, position=obstacle.body.position, half_size=self.obj_size * 0.5)
        for goal in self.goals:
            if self.goal_sprite is not None:
                draw_sprite(self.screen, image=self.goal_sprite, position=goal.body.position, half_size=self.obj_size * 0.5)
            else:
                draw_rectangle(self.screen, color=self.goal_color, position=goal.body.position, half_size=self.obj_size * 0.5)
        for block in self.blocks:
            if self.block_sprite is not None:
                draw_sprite(self.screen, image=self.block_sprite, position=block.body.position, half_size=self.obj_size * 0.5)
            else:
                draw_rectangle(self.screen, color=self.block_color, position=block.body.position, half_size=self.obj_size * 0.5)
        for player in self.players:
            if self.player_sprite is not None:
                draw_sprite(self.screen, image=self.player_sprite, position=player.body.position, half_size=self.obj_size * 0.5)
            else:
                draw_circle(self.screen, color=self.player_color, position=player.body.position, radius=self.obj_size * 0.5)

    @staticmethod
    def _get_distance(shape, shapes, method='e'):
        min_distance = 1000000
        for other_shape in shapes:
            if method == 'e':
                distance = euclidean_distance(position1=shape.body.position, position2=other_shape.body.position)
            else:
                distance = manhattan_distance(position1=shape.body.position, position2=other_shape.body.position)

            if distance < min_distance:
                min_distance = distance

        return min_distance

    def _get_info(self):
        return self.overlaps

    def _get_observation(self):
        self._draw()
        image = pygame.surfarray.array3d(self.screen)
        image = np.swapaxes(image, 0, 1)
        return image

    def _get_reward(self):
        # remove movement penalty to agent
        total_reward = 0
        # total_reward = self.movement_penalty

        score = self._get_score()
        total_reward += (self.score - score)
        self.score = score

        # move 0.9 to xml file as well
        successes = len([v for (_, _), v in self.overlaps.items() if v >= .9])
        if successes > self.successes:
            total_reward += self.box_on_target
        elif successes < self.successes:
            total_reward += self.box_off_target
        self.successes = successes

        self._check_complete()
        if self.level_complete:
            total_reward += self.all_boxes_on_target
            return total_reward

        return total_reward

    def _get_score(self):
        total = 0
        for block in self.blocks:
            block_min = self._get_distance(block, self.goals)
            total += (block_min * 1)
        return total

    def _get_terminal(self):
        return self.level_complete

    def _init_space(self):
        self.space = pymunk.Space()
        self.space.gravity = self.gravity
        self.space.damping = self.damping

        block_goal = self.space.add_collision_handler(self.block_collision, self.goal_collision)
        block_goal.begin = self._begin
        block_goal.pre_solve = self._pre_solve
        block_goal.post_solve = self._post_solve
        block_goal.separate = self._separate

    def _init_window(self):
        pygame.init()
        self.screen = pygame.Surface(self.world_metrics)
        self.screen.fill(self.ground_color)

    def _play_step(self, force=None):
        if force is None:
            force = Vec2d(0.0, 0.0)
        self.player.force = force

        for i in range(self.steps):
            if i < self.steps - 1:
                self.space.damping = 1.0
            else:
                self.space.damping = self.damping

            for shape in self.space.shapes:
                self.space.add_post_step_callback(self._post_step_callback, shape)
            self.space.step(self.dt)

    def _post_solve(self, arbiter, _, __):
        if arbiter.shapes[0].collision_type == self.goal_collision:
            goal = arbiter.shapes[0]
            block = arbiter.shapes[1]
        else:
            goal = arbiter.shapes[1]
            block = arbiter.shapes[0]

        goal_center = goal.body.position
        block_center = block.body.position

        center_difference = Vec2d(abs(goal_center.x - block_center.x), abs(goal_center.y - block_center.y))
        filled_area_sizes = Vec2d(self.obj_size, self.obj_size) - center_difference
        filled_area = abs(filled_area_sizes.x * filled_area_sizes.y)
        filled_area_percentage = filled_area / goal.area

        self.overlaps[(goal.body, block.body)] = filled_area_percentage

    @staticmethod
    def _post_step_callback(space, key):
        key.body.angle = 0
        space.reindex_shapes_for_body(key.body)

    def _pre_solve(self, arbiter, _, __):
        if arbiter.shapes[0].collision_type == self.goal_collision:
            goal = arbiter.shapes[0]
            block = arbiter.shapes[1]
        else:
            goal = arbiter.shapes[1]
            block = arbiter.shapes[0]

        goal_center = goal.body.position
        block_center = block.body.position

        center_difference = Vec2d(abs(goal_center.x - block_center.x), abs(goal_center.y - block_center.y))
        filled_area_sizes = Vec2d(self.obj_size, self.obj_size) - center_difference
        filled_area = abs(filled_area_sizes.x * filled_area_sizes.y)
        filled_area_percentage = filled_area / goal.area

        self.overlaps[(goal.body, block.body)] = filled_area_percentage
        return True

    def _recreate(self):
        self._create_objects()

        self.obstacles = list(filter(lambda shape: shape.collision_type == self.obstacle_collision, self.space.shapes))
        self.players = list(filter(lambda shape: shape.collision_type == self.player_collision, self.space.shapes))
        self.blocks = list(filter(lambda shape: shape.collision_type == self.block_collision, self.space.shapes))
        self.goals = list(filter(lambda shape: shape.collision_type == self.goal_collision, self.space.shapes))

        self.player = self.players[0].body

        self.score = self._get_score()
        self.successes = 0

    def _separate(self, arbiter, _, __):
        if arbiter.shapes[0].collision_type == self.goal_collision:
            goal = arbiter.shapes[0]
            block = arbiter.shapes[1]
        else:
            goal = arbiter.shapes[1]
            block = arbiter.shapes[0]

        if (goal.body, block.body) in self.overlaps.keys():
            del self.overlaps[(goal.body, block.body)]

    def close(self):
        self._clear()
        cv2.destroyAllWindows()
        pygame.quit()

    def render(self, mode='human'):
        image = pygame.surfarray.array3d(self.screen)
        image = np.swapaxes(image, 0, 1)
        cv2.imshow('state', cv2.cvtColor(image, cv2.COLOR_RGB2BGR))
        cv2.waitKey(1)

    def reset(self):
        self._clear()
        self.layout = self.new_layout()
        self._recreate()
        self._play_step()
        return self._get_observation()

    def seed(self, seed=None):
        pass

    def step(self, action):
        try:
            x, y = action
            vector = Vec2d(x, y)
            if vector.length > 1.0:
                vector = vector.normalized
        except TypeError:
            vector = Vec2d(0.0, 0.0)
        self._play_step(force=vector * self.force)
        return self._get_observation(), self._get_reward(), self._get_terminal(), self._get_info()

    def goal(self):
        screen = pygame.Surface(self.world_metrics)
        screen.fill(self.ground_color)

        for obstacle in self.obstacles:
            if self.obstacle_sprite is not None:
                draw_sprite(
                    screen,
                    image=self.obstacle_sprite,
                    position=obstacle.body.position,
                    half_size=self.obj_size * 0.5
                )
            else:
                draw_rectangle(
                    screen,
                    color=self.obstacle_color,
                    position=obstacle.body.position,
                    half_size=self.obj_size * 0.5
                )

        player = self.players[0]
        for goal in self.goals:
            if self.goal_sprite is not None:
                draw_sprite(
                    screen,
                    image=self.goal_sprite,
                    position=goal.body.position,
                    half_size=self.obj_size * 0.5
                )
            else:
                draw_rectangle(
                    screen,
                    color=self.goal_color,
                    position=goal.body.position,
                    half_size=self.obj_size * 0.5
                )

            if self.block_sprite is not None:
                draw_sprite(
                    screen,
                    image=self.block_sprite,
                    position=player.body.position - (player.body.position - goal.body.position) * 0.95,
                    half_size=self.obj_size * 0.5
                )
            else:
                draw_rectangle(
                    screen,
                    color=self.block_color,
                    position=player.body.position - (player.body.position - goal.body.position) * 0.95,
                    half_size=self.obj_size * 0.5
                )

        player = self.players[0]
        block = self.blocks[0]
        if self.player_sprite is not None:
            draw_sprite(
                screen, image=self.player_sprite, position=player.body.position - (player.body.position - block.body.position) * 0.9,
                half_size=self.obj_size * 0.5
            )
        else:
            draw_circle(
                screen, color=self.player_color, position=player.body.position - (player.body.position - block.body.position) * 0.9,
                radius=self.obj_size * 0.5
            )

        image = pygame.surfarray.array3d(screen)
        image = np.swapaxes(image, 0, 1)
        return image



class Graph:
    def __init__(self, edges, n):
        self.adj_list = [[] for _ in range(n)]

        for (src, dest) in edges:
            self.adj_list[src].append(dest)
            self.adj_list[dest].append(src)



import functools
import inspect
import json

from joatmon import context
from joatmon.utility import (
    JSONEncoder,
    to_enumerable,
    to_hash
)


def cached(cache, duration):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            cache_value = context.get_value(cache)

            key = to_hash(func, *args, **kwargs)
            if (value := await cache_value.get(key)) is None:
                result = await func(*args, **kwargs)
                await cache_value.add(key, json.dumps(to_enumerable(result), cls=JSONEncoder), duration)
            else:
                result = to_enumerable(json.loads(value))

            return result

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator


def remove(cache, regex):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            cache_value = context.get_value(cache)
            await cache_value.remove(regex)
            return await func(*args, **kwargs)

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator



__all__ = ['File']

from joatmon.system.lock import RWLock


class File:
    def __init__(self, path):
        self.path = path
        self.lock = RWLock()

    def read(self):
        with self.lock.r_locked():
            with open(str(self.path), 'r') as file:
                return file.read()

    def write(self, data=''):
        with self.lock.w_locked():
            with open(str(self.path), 'w') as file:
                file.write(data + '\n')



def normal(param, loc=0.0, scale=1.0):
    import numpy as np
    param._data = np.random.normal(loc, scale, size=param.shape)


def uniform(param, low=-1.0, high=1.0):
    import numpy as np
    param._data = np.random.uniform(low, high, size=param.shape)



import cv2
import numpy as np
import pygame

__all__ = ['CubeEnv']

from joatmon.game.core import CoreEnv

cube_colors = ['blue', 'red', 'yellow', 'green', 'orange', 'white']
cube_colors = list(map(lambda x: pygame.color.THECOLORS[x], cube_colors))

screen_size = 400


def flip(pos):
    return pos[0], -pos[1] + screen_size


def project_points(points, q, view, vertical=(0, 1, 0)):
    points = np.asarray(points)
    view = np.asarray(view)

    xdir = np.cross(vertical, view).astype(float)

    if np.all(xdir == 0):
        raise ValueError("vertical is parallel to v")

    xdir /= np.sqrt(np.dot(xdir, xdir))

    ydir = np.cross(view, xdir)
    ydir /= np.sqrt(np.dot(ydir, ydir))

    v2 = np.dot(view, view)
    zdir = view / np.sqrt(v2)

    r = q.as_rotation_matrix()
    rpts = np.dot(points, r.T)

    dpoint = rpts - view
    dpoint_view = np.dot(dpoint, view).reshape(dpoint.shape[:-1] + (1,))
    dproj = -dpoint * v2 / dpoint_view

    trans = list(range(1, dproj.ndim)) + [0]
    return np.array(
        [
            np.dot(dproj, xdir), np.dot(dproj, ydir), -np.dot(dpoint, zdir)
        ]
    ).transpose(trans)


def resize(pos):
    return pos * (screen_size // 4) + (screen_size // 2)


class Quaternion:
    def __init__(self, x):
        self.x = np.asarray(x, dtype=float)

    def __mul__(self, other):
        sxr = self.x.reshape(self.x.shape[:-1] + (4, 1))
        oxr = other.x.reshape(other.x.shape[:-1] + (1, 4))

        prod = sxr * oxr
        return_shape = prod.shape[:-1]
        prod = prod.reshape((-1, 4, 4)).transpose((1, 2, 0))

        ret = np.array(
            [
                (prod[0, 0] - prod[1, 1] - prod[2, 2] - prod[3, 3]),
                (prod[0, 1] + prod[1, 0] + prod[2, 3] - prod[3, 2]),
                (prod[0, 2] - prod[1, 3] + prod[2, 0] + prod[3, 1]),
                (prod[0, 3] + prod[1, 2] - prod[2, 1] + prod[3, 0])
            ], dtype=np.float, order='F'
        ).T
        return self.__class__(ret.reshape(return_shape))

    def __repr__(self):
        return "Quaternion:\n" + self.x.__repr__()

    def as_rotation_matrix(self):
        v, theta = self.as_v_theta()

        shape = theta.shape
        theta = theta.reshape(-1)
        v = v.reshape(-1, 3).T
        c = np.cos(theta)
        s = np.sin(theta)

        mat = np.array(
            [
                [v[0] * v[0] * (1 - c) + c, v[0] * v[1] * (1 - c) - v[2] * s, v[0] * v[2] * (1 - c) + v[1] * s],
                [v[1] * v[0] * (1 - c) + v[2] * s, v[1] * v[1] * (1 - c) + c, v[1] * v[2] * (1 - c) - v[0] * s],
                [v[2] * v[0] * (1 - c) - v[1] * s, v[2] * v[1] * (1 - c) + v[0] * s, v[2] * v[2] * (1 - c) + c]
            ], order='F'
        ).T
        return mat.reshape(shape + (3, 3))

    def as_v_theta(self):
        x = self.x.reshape((-1, 4)).T

        norm = np.sqrt((x ** 2).sum(0))
        theta = 2 * np.arccos(x[0] / norm)

        v = np.array(x[1:], order='F')
        v /= np.sqrt(np.sum(v ** 2, 0))

        v = v.T.reshape(self.x.shape[:-1] + (3,))
        theta = theta.reshape(self.x.shape[:-1])

        return v, theta

    @classmethod
    def from_v_theta(cls, v, theta):
        theta = np.asarray(theta)
        v = np.asarray(v)
        s = np.sin(0.5 * theta)
        c = np.cos(0.5 * theta)

        v = v * s / np.sqrt(np.sum(v * v, -1))
        x_shape = v.shape[:-1] + (4,)

        x = np.ones(x_shape).reshape(-1, 4)
        x[:, 0] = c.ravel()
        x[:, 1:] = v.reshape(-1, 3)
        x = x.reshape(x_shape)

        return cls(x)

    def rotate(self, points):
        m = self.as_rotation_matrix()
        return np.dot(points, m.T)


class Sticker:
    def __init__(self, color):
        self.color = color

    def draw(self, screen, points, view, rotation, vertical):
        points = project_points(points, rotation, view, vertical)[:, :2]
        points = list(map(resize, points))
        points = list(map(flip, points))
        pygame.draw.polygon(screen, cube_colors[self.color], points)
        pygame.draw.polygon(screen, pygame.color.THECOLORS['black'], points, 1)


class Face:
    def __init__(self, n, color, top_left, increment):
        self.n = n
        self.top_left = top_left
        self.increment = increment

        self.stickers = [[Sticker(color) for _ in range(n)] for _ in range(n)]

    def draw(self, screen, view, rotation, vertical):
        for i, row in enumerate(self.stickers):
            for j, sticker in enumerate(row):
                sticker.draw(
                    screen, (self.top_left + self.increment[0] * (i + 0) + self.increment[1] * (j + 0),
                             self.top_left + self.increment[0] * (i + 0) + self.increment[1] * (j + 1),
                             self.top_left + self.increment[0] * (i + 1) + self.increment[1] * (j + 1),
                             self.top_left + self.increment[0] * (i + 1) + self.increment[1] * (j + 0),
                             self.top_left + self.increment[0] * (i + 0) + self.increment[1] * (j + 0)), view, rotation, vertical
                )

    def rotate(self, times):
        for _ in range(times):
            self.rot90()

    def rotate_layer(self, layer):
        for ind in range(layer, self.n - layer - 1):
            tmp = self.stickers[layer][ind].color
            self.stickers[layer][ind].color = self.stickers[self.n - layer - ind - 1][layer].color
            self.stickers[self.n - layer - ind - 1][layer].color = self.stickers[self.n - layer - 1][self.n - layer - ind - 1].color
            self.stickers[self.n - layer - 1][self.n - layer - ind - 1].color = self.stickers[layer + ind][self.n - layer - 1].color
            self.stickers[layer + ind][self.n - layer - 1].color = tmp

    def rot90(self):
        for layer in range(self.n // 2):
            self.rotate_layer(layer)


class Cube:
    # blue left, red front, yellow top
    # green right, orange back, white bottom
    def __init__(self, n=3):
        self.n = n
        self.faces = {
            'L': Face(
                n, 0, np.asarray([-1, +1, -1]).astype('float32'),
                (np.asarray([0, -1, 0]).astype('float32') * (2.0 / n), np.asarray([0, 0, +1]).astype('float32') * (2.0 / n))
            ),
            'F': Face(
                n, 1, np.asarray([-1, +1, +1]).astype('float32'),
                (np.asarray([0, -1, 0]).astype('float32') * (2.0 / n), np.asarray([+1, 0, 0]).astype('float32') * (2.0 / n))
            ),
            'U': Face(
                n, 2, np.asarray([-1, +1, -1]).astype('float32'),
                (np.asarray([0, 0, +1]).astype('float32') * (2.0 / n), np.asarray([+1, 0, 0]).astype('float32') * (2.0 / n))
            ),
            'R': Face(
                n, 3, np.asarray([+1, +1, +1]).astype('float32'),
                (np.asarray([0, -1, 0]).astype('float32') * (2.0 / n), np.asarray([0, 0, -1]).astype('float32') * (2.0 / n))
            ),
            'B': Face(
                n, 4, np.asarray([+1, +1, -1]).astype('float32'),
                (np.asarray([0, -1, 0]).astype('float32') * (2.0 / n), np.asarray([-1, 0, 0]).astype('float32') * (2.0 / n))
            ),
            'D': Face(
                n, 5, np.asarray([-1, -1, +1]).astype('float32'),
                (np.asarray([0, 0, -1]).astype('float32') * (2.0 / n), np.asarray([+1, 0, 0]).astype('float32') * (2.0 / n))
            )
        }
        self.order = ['D', 'B', 'L', 'R', 'F', 'U']

        self.front = pygame.Surface((screen_size, screen_size))
        self.back = pygame.Surface((screen_size, screen_size))

        self.screen = pygame.Surface((screen_size, screen_size))

        self.view = (0, 0, 15)
        self.rotation = Quaternion.from_v_theta((1, -1, 0), -np.pi / 6)
        self.vertical = [0, 1, 0]

    def draw(self):
        self.back.fill(pygame.color.THECOLORS['black'])
        for face_name in self.order[:3]:
            self.faces[face_name].draw(self.back, list(map(lambda x: -x, self.view)), self.rotation, list(map(lambda x: -x, self.vertical)))
        cv2.imshow('back', cv2.cvtColor(np.swapaxes(pygame.surfarray.array3d(self.back), 0, 1), cv2.COLOR_RGB2BGR))

        self.front.fill(pygame.color.THECOLORS['black'])
        for face_name in self.order[3:]:
            self.faces[face_name].draw(self.front, self.view, self.rotation, self.vertical)
        cv2.imshow('front', cv2.cvtColor(np.swapaxes(pygame.surfarray.array3d(self.front), 0, 1), cv2.COLOR_RGB2BGR))

        cv2.waitKey(0)

    def swap_faces(self, faces):
        for row in range(self.n):
            for col in range(self.n):
                tmp = self.faces[faces[0]].stickers[row][col].color
                self.faces[faces[0]].stickers[row][col].color = self.faces[faces[1]].stickers[row][col].color
                self.faces[faces[1]].stickers[row][col].color = self.faces[faces[2]].stickers[row][col].color
                self.faces[faces[2]].stickers[row][col].color = self.faces[faces[3]].stickers[row][col].color
                self.faces[faces[3]].stickers[row][col].color = tmp

    def swap_layers(self, faces, layer):
        pass

    def u(self):
        self.faces['U'].rotate(1)
        for ind in range(self.n):
            tmp = self.faces['F'].stickers[0][ind].color
            self.faces['F'].stickers[0][ind].color = self.faces['R'].stickers[0][ind].color
            self.faces['R'].stickers[0][ind].color = self.faces['B'].stickers[0][ind].color
            self.faces['B'].stickers[0][ind].color = self.faces['L'].stickers[0][ind].color
            self.faces['L'].stickers[0][ind].color = tmp

    def x(self):
        self.swap_faces(['F', 'D', 'B', 'U'])
        self.faces['L'].rotate(1)
        self.faces['R'].rotate(3)

    def y(self):
        self.swap_faces(['F', 'R', 'B', 'L'])
        self.faces['U'].rotate(1)
        self.faces['D'].rotate(3)

    def z(self):
        self.swap_faces(['R', 'U', 'L', 'D'])
        self.faces['F'].rotate(1)
        self.faces['B'].rotate(3)


class CubeEnv(CoreEnv):
    def __init__(self):
        super(CubeEnv, self).__init__()

    def close(self):
        pass

    def render(self, mode: str = 'human'):
        pass

    def reset(self):
        pass

    def seed(self, seed=None):
        pass

    def step(self, action):
        pass


if __name__ == '__main__':
    cube = Cube()
    cube.x()
    cube.draw()

    # points = project_points(
    #     [
    #         (0, 0, -150),
    #         (150, 0, -150),
    #         (150, -150, -150),
    #         (0, -150, -150),
    #         (0, 0, -150)
    #     ],
    #     Quaternion.from_v_theta((1, -1, 0), -np.pi / 6),
    #     (0, 0, 200),
    #     [0, 1, 0]
    # )[:, :2]
    # print(points)



import os

__all__ = ['Path']


class Path:
    def __init__(self, path=''):
        if path is None or path == '':
            path = os.getcwd()
        self.path = path

    def __str__(self):
        return self.path

    def __repr__(self):
        return str(self)

    def __truediv__(self, other):
        if isinstance(other, Path):
            other = other.path

        if other is None or other == '' or other == '/':
            other = '.'

        if other == '.':
            return self

        return Path(os.path.join(self.path, other))

    def __itruediv__(self, other):
        new = self / other
        self.path = new.path

        return self

    def __abs__(self):
        if not self.path.startswith('/'):
            cwd = os.getcwd()
            new = Path(cwd) / self

            return new

        return self

    def exists(self):
        return os.path.exists(self.path)

    def isdir(self):
        return os.path.isdir(self.path)

    def isfile(self):
        return os.path.isfile(self.path)

    def mkdir(self):
        return os.mkdir(self.path)

    def list(self):
        return os.listdir(self.path)



from joatmon import context
from joatmon.plugin.core import Plugin


class UserPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class TokenPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class IssuerPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class LanguagePlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class IPPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class InCasePlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class OutCasePlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class JsonPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class ArgsPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class FormPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class HeadersPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)


class CookiesPlugin(Plugin):
    def __init__(self, key):
        self.key = key

    def set(self, value):
        context.set_value(self.key, value)

    def get(self):
        return context.get_value(self.key)



from joatmon.nn import functional as f
from joatmon.nn.core import Module

__all__ = ['ReLU']


class ReLU(Module):
    def __init__(self, alpha=0):
        super(ReLU, self).__init__()

        self.alpha = alpha

    def forward(self, inp):
        return f.relu(
            inp=inp,
            alpha=self.alpha
        )



from joatmon.orm.field import get_converter
from joatmon.orm.meta import Meta
from joatmon.serializable import Serializable


class Document(Serializable):  # need to have copy and deepcopy functions as well
    __metaclass__ = Meta

    def __init__(self, **kwargs):
        super(Document, self).__init__(**kwargs)

        for name, field in self.__metaclass__.fields(self.__metaclass__).items():
            if name not in kwargs:
                setattr(self, name, None)

    def __getattr__(self, item):
        return self.__dict__.get(item, None)

    def __setattr__(self, key, value):
        if key not in self.__metaclass__.fields(self.__metaclass__).keys():
            self.__dict__[key] = value
            return

        field = self.__metaclass__.fields(self.__metaclass__)[key]
        # self.__dict__[key] = get_converter(field.dtype)(value)
        self.__dict__[key] = get_converter(field)(value)

    def __len__(self):
        return len(self.__dict__.keys())

    def __iter__(self):
        for k in self.__dict__.keys():
            yield k, getattr(self, k, None)

    def __getitem__(self, item):
        return getattr(self, item, None)

    def keys(self):
        return list(self.__dict__.keys())

    def values(self):
        return list(map(lambda x: getattr(self, x, None), self.__dict__.keys()))

    def validate(self):
        if not self.__metaclass__.structured and not self.__metaclass__.force:
            raise ValueError('unstructured document cannot be validated')

        ret = {}
        for name, field in self.__metaclass__.fields(self.__metaclass__).items():
            value = getattr(self, name, None)

            default_value = field.default()

            if value is None and not field.nullable:
                setattr(self, name, default_value)

            value = getattr(self, name, None)

            if value is None and not field.nullable:
                raise ValueError(f'field {name} is not nullable')

            if isinstance(field.dtype, (tuple, list)):
                if ((value is not None and field.nullable) or not field.nullable) and not isinstance(value, field.dtype):
                    raise ValueError(
                        f'field {name} has to be one of the following {field.dtype} not {type(value).__name__}'
                    )
            else:
                if ((value is not None and field.nullable) or not field.nullable) and type(value) is not field.dtype:
                    raise ValueError(f'field {name} has to be type {field.dtype} not {type(value).__name__}')

            constraints = self.__metaclass__.constraints(self.__metaclass__).values()
            constraints = list(filter(lambda x: x.field == name, constraints))

            if not field.nullable and constraints is not None:
                for constraint in constraints:
                    constraint.check(value)

            ret[name] = getattr(self, name, None)

        return ret


def create_new_type(meta, subclasses):
    return type(meta.__collection__, subclasses, {'__metaclass__': meta})



import copy
import os

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim

from joatmon.ai.models.core import CoreModel
from joatmon.ai.network.reinforcement.hybrid.td3 import (
    TD3Actor,
    TD3Critic
)
from joatmon.ai.utility import (
    load,
    save,
    to_numpy,
    to_tensor
)

__all__ = ['TD3Model']


class TD3Model(CoreModel):
    """
    Twin Delayed Deep Deterministic Policy Gradient

    # Arguments
        actor_model (`keras.nn.Model` instance): See [Model](#) for details.
        critic_model (`keras.nn.Model` instance): See [Model](#) for details.
        optimizer (`keras.optimizers.Optimizer` instance):
        See [Optimizer](#) for details.
        action_inp (`keras.layers.Input` / `keras.layers.InputLayer` instance):
        See [Input](#) for details.
        tau (float): tau.
        gamma (float): gamma.
    """

    def __init__(self, lr=1e-3, tau=1e-4, gamma=0.99, in_features=1, out_features=1):
        super(TD3Model, self).__init__()

        self.actor_local = TD3Actor(in_features, out_features)
        self.actor_target = copy.deepcopy(self.actor_local)
        self.hardupdate('actor')

        self.critic_local_1 = TD3Critic(in_features, out_features)
        self.critic_target_1 = copy.deepcopy(self.critic_local_1)
        self.hardupdate('critic_1')

        self.critic_local_2 = TD3Critic(in_features, out_features)
        self.critic_target_2 = copy.deepcopy(self.critic_local_2)
        self.hardupdate('critic_2')

        self.lr = lr
        self.tau = tau
        self.gamma = gamma

        self.actor_optimizer = optim.Adam(self.actor_local.parameters(), lr=self.lr)

        self.critic_optimizer = optim.Adam(list(self.critic_local_1.parameters()) + list(self.critic_local_2.parameters()), lr=self.lr)
        self.critic_loss = nn.SmoothL1Loss()

    def load(self, path=''):
        load(self.actor_local, os.path.join(path, 'actor_local'))
        load(self.actor_target, os.path.join(path, 'actor_target'))
        load(self.critic_local_1, os.path.join(path, 'critic_local_1'))
        load(self.critic_target_1, os.path.join(path, 'critic_target_1'))
        load(self.critic_local_2, os.path.join(path, 'critic_local_2'))
        load(self.critic_target_2, os.path.join(path, 'critic_target_2'))

    def save(self, path=''):
        save(self.actor_local, os.path.join(path, 'actor_local'))
        save(self.actor_target, os.path.join(path, 'actor_target'))
        save(self.critic_local_1, os.path.join(path, 'critic_local_1'))
        save(self.critic_target_1, os.path.join(path, 'critic_target_1'))
        save(self.critic_local_2, os.path.join(path, 'critic_local_2'))
        save(self.critic_target_2, os.path.join(path, 'critic_target_2'))

    def initialize(self, w_init=None, b_init=None):
        for module in self.actor_local.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)
        for module in self.actor_target.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)
        for module in self.critic_local_1.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)
        for module in self.critic_target_1.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)
        for module in self.critic_local_2.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)
        for module in self.critic_target_2.modules():
            if isinstance(module, (torch.nn.Conv2d, torch.nn.BatchNorm2d, torch.nn.Linear)):  # batch norm will be different
                if w_init is not None:
                    torch.nn.init.kaiming_normal(module.weight)
                if b_init is not None:  # bias init will be different
                    torch.nn.init.kaiming_normal(module.bias)

    def softupdate(self, network: str):
        if network == 'actor':
            for target_param, param in zip(self.actor_target.parameters(), self.actor_local.parameters()):
                target_param.detach_()
                target_param.copy_(target_param * (1.0 - self.tau) + param * self.tau)
        elif network == 'critic_1':
            for target_param, param in zip(self.critic_target_1.parameters(), self.critic_local_1.parameters()):
                target_param.detach_()
                target_param.copy_(target_param * (1.0 - self.tau) + param * self.tau)
        elif network == 'critic_2':
            for target_param, param in zip(self.critic_target_2.parameters(), self.critic_local_2.parameters()):
                target_param.detach_()
                target_param.copy_(target_param * (1.0 - self.tau) + param * self.tau)
        else:
            raise ValueError

    def hardupdate(self, network: str):
        if network == 'actor':
            for target_param, param in zip(self.actor_target.parameters(), self.actor_local.parameters()):
                target_param.detach_()
                target_param.copy_(param)
        elif network == 'critic_1':
            for target_param, param in zip(self.critic_target_1.parameters(), self.critic_local_1.parameters()):
                target_param.detach_()
                target_param.copy_(param)
        elif network == 'critic_2':
            for target_param, param in zip(self.critic_target_2.parameters(), self.critic_local_2.parameters()):
                target_param.detach_()
                target_param.copy_(param)
        else:
            raise ValueError

    def predict(self, state=None):
        self.actor_local.eval()
        return to_numpy(self.actor_local(to_tensor(state))).flatten()

    def train(self, batch=None, update_target=True):
        self.actor_local.train()

        states, actions, rewards, next_states, terminals = batch

        states = to_tensor(states)
        actions = to_tensor(actions)
        rewards = to_tensor(rewards).unsqueeze(-1)
        next_states = to_tensor(next_states)
        terminals = to_tensor(terminals).unsqueeze(-1)

        a_next = self.actor_target(next_states)
        q_next_1 = self.critic_target_1(next_states, a_next)
        q_next_2 = self.critic_target_2(next_states, a_next)
        q_next = torch.min(q_next_1, q_next_2)

        q_next = self.gamma * q_next * (1 - terminals)
        q_next.add_(rewards)
        q_next = q_next.detach()

        q_1 = self.critic_local_1(states, actions)
        q_2 = self.critic_local_2(states, actions)

        critic_1_loss = F.mse_loss(q_1, q_next)
        critic_2_loss = F.mse_loss(q_2, q_next)

        critic_loss = critic_1_loss + critic_2_loss
        self.critic_optimizer.zero_grad()
        critic_loss.backward()
        self.critic_optimizer.step()

        # might need to update with policy update frequency
        a = self.actor_local(states)

        actor_loss = -self.critic_local_1(states, a).mean()
        self.actor_optimizer.zero_grad()
        actor_loss.backward()
        self.actor_optimizer.step()

        if update_target:
            self.softupdate('actor')
            self.softupdate('critic_1')
            self.softupdate('critic_2')

        return [actor_loss.item()] + [critic_loss.item()]

    def evaluate(self):
        pass



from enum import Enum


class Arithmetic(Enum):
    add = "+"
    sub = "-"
    mul = "*"
    div = "/"


class Comparator(Enum):
    pass


class Equality(Comparator):
    eq = "="
    ne = "<>"
    gt = ">"
    gte = ">="
    lt = "<"
    lte = "<="


class Matching(Comparator):
    not_like = " NOT LIKE "
    like = " LIKE "


class Boolean(Comparator):
    and_ = "AND"
    or_ = "OR"
    xor_ = "XOR"
    true = "TRUE"
    false = "FALSE"


class Order(Enum):
    asc = "ASC"
    desc = "DESC"


class JoinType(Enum):
    inner = ""
    left = "LEFT"
    right = "RIGHT"
    outer = "FULL OUTER"
    left_outer = "LEFT OUTER"
    right_outer = "RIGHT OUTER"
    full_outer = "FULL OUTER"
    cross = "CROSS"


class DatePart(Enum):
    year = "YEAR"
    quarter = "QUARTER"
    month = "MONTH"
    week = "WEEK"
    day = "DAY"
    hour = "HOUR"
    minute = "MINUTE"
    second = "SECOND"
    microsecond = "MICROSECOND"


class Dialects(Enum):
    MSSQL = "mssql"
    MYSQL = "mysql"
    POSTGRESQL = "postgressql"
    SQLLITE = "sqllite"
    MONGO = "mongo"


class Node:
    def __init__(self, alias):
        self.alias = alias

    def as_(self, alias):
        self.alias = alias
        return self


class Term(Node):
    def __init__(self, alias):
        super(Term, self).__init__(alias)

    def __eq__(self, other):
        return BasicCriteria(Equality.eq, self, other)

    def __ne__(self, other):
        return BasicCriteria(Equality.ne, self, other)

    def __gt__(self, other):
        return BasicCriteria(Equality.gt, self, other)

    def __ge__(self, other):
        return BasicCriteria(Equality.gte, self, other)

    def __lt__(self, other):
        return BasicCriteria(Equality.lt, self, other)

    def __le__(self, other):
        return BasicCriteria(Equality.lte, self, other)

    def __sub__(self, other):
        return ArithmeticExpression(Arithmetic.sub, self, other)

    def __add__(self, other):
        return ArithmeticExpression(Arithmetic.add, self, other)

    def __mul__(self, other):
        return ArithmeticExpression(Arithmetic.mul, self, other)

    def __truediv__(self, other):
        return ArithmeticExpression(Arithmetic.div, self, other)

    def build(self, dialect, depth=0):
        raise NotImplementedError


class ValueWrapper:
    def __init__(self, value):
        self.value = value

    def build(self, dialect, depth=0):
        if self.value is None:
            return 'null'
        elif self.value is True:
            return 'true'
        elif self.value is False:
            return 'false'
        elif isinstance(self.value, str):
            return f'\'{self.value}\''
        return str(self.value)


class Criterion(Term):
    def __and__(self, other):
        return ComplexCriteria(Boolean.and_, self, other)

    def __or__(self, other):
        return ComplexCriteria(Boolean.or_, self, other)

    def __xor__(self, other):
        ...

    @staticmethod
    def any():
        ...

    @staticmethod
    def all():
        ...


class BasicCriteria(Criterion):
    def __init__(self, comparator: Comparator, left: Term, right: Term, alias=None):
        super(BasicCriteria, self).__init__(alias)

        self.comparator = comparator
        self.left = left
        if right is None:
            right = ValueWrapper(right)
        if isinstance(right, (bool, int, float, str)):
            right = ValueWrapper(right)
        self.right = right

    def build(self, dialect, depth=0):
        if dialect == Dialects.MSSQL:
            ret = ''
            if self.comparator == Equality.eq:
                return f'{self.left.build(dialect)} = {self.right.build(dialect)}'
            if self.comparator == Equality.ne:
                return f'{self.left.build(dialect)} != {self.right.build(dialect)}'
            if self.comparator == Equality.gt:
                return f'{self.left.build(dialect)} > {self.right.build(dialect)}'
            if self.comparator == Equality.gte:
                return f'{self.left.build(dialect)} >= {self.right.build(dialect)}'
            if self.comparator == Equality.lt:
                return f'{self.left.build(dialect)} < {self.right.build(dialect)}'
            if self.comparator == Equality.lte:
                return f'{self.left.build(dialect)} <= {self.right.build(dialect)}'
            return ret
        if dialect == Dialects.POSTGRESQL:
            ret = ''
            if self.comparator == Equality.eq:
                if isinstance(self.right, ValueWrapper) and self.right.value is None:
                    ret += f'{self.left.build(dialect)} is {self.right.build(dialect)}'
                else:
                    ret += f'{self.left.build(dialect)} = {self.right.build(dialect)}'
            if self.comparator == Equality.ne:
                if isinstance(self.right, ValueWrapper) and self.right.value is None:
                    ret += f'{self.left.build(dialect)} is not {self.right.build(dialect)}'
                else:
                    ret += f'{self.left.build(dialect)} != {self.right.build(dialect)}'
            if self.comparator == Equality.gt:
                ret += f'{self.left.build(dialect)} > {self.right.build(dialect)}'
            if self.comparator == Equality.gte:
                ret += f'{self.left.build(dialect)} >= {self.right.build(dialect)}'
            if self.comparator == Equality.lt:
                ret += f'{self.left.build(dialect)} < {self.right.build(dialect)}'
            if self.comparator == Equality.lte:
                ret += f'{self.left.build(dialect)} <= {self.right.build(dialect)}'
            return ret if self.alias is None else f'{ret} as {self.alias}'
        if dialect == Dialects.MONGO:
            ret = {}
            if self.comparator == Equality.eq:
                ret[self.left.build(dialect)] = {'$eq': self.right.build(dialect)}
            print(ret)
            return ret


class ComplexCriteria(Criterion):
    def __init__(self, comparator: Comparator, left: Term, right: Term, alias=None):
        super(ComplexCriteria, self).__init__(alias)

        self.comparator = comparator
        self.left = left
        self.right = right

    def build(self, dialect, depth=0):
        if dialect == Dialects.MSSQL:
            ret = ''
            if self.comparator == Boolean.and_:
                return f'{self.left.build(dialect)} and \n{self.right.build(dialect)}'
            if self.comparator == Boolean.or_:
                return f'{self.left.build(dialect)} or \n{self.right.build(dialect)}'
            return ret
        if dialect == Dialects.POSTGRESQL:
            ret = ''
            if self.comparator == Boolean.and_:
                ret += f'{self.left.build(dialect)} and \n{self.right.build(dialect)}'
            if self.comparator == Boolean.or_:
                ret += f'{self.left.build(dialect)} or \n{self.right.build(dialect)}'
            return ret if self.alias is None else f'{ret} as {self.alias}'
        if dialect == Dialects.MONGO:
            ret = {}
            if self.comparator == Boolean.and_:
                ret['$and'] = [self.left.build(dialect), self.right.build(dialect)]
            print(ret)
            return ret


class ArithmeticExpression(Term):
    def __init__(self, operator: Arithmetic, left, right, alias=None):
        super(ArithmeticExpression, self).__init__(alias)

        self.operator = operator
        self.left = left
        if isinstance(right, (bool, int, float, str)):
            right = ValueWrapper(right)
        self.right = right

    def build(self, dialect, depth=0):
        if dialect == Dialects.MSSQL:
            ret = '('
            if self.operator == Arithmetic.add:
                ret += f'{self.left.build(dialect)} + {self.right.build(dialect)}'
            if self.operator == Arithmetic.sub:
                ret += f'{self.left.build(dialect)} - {self.right.build(dialect)}'
            if self.operator == Arithmetic.mul:
                ret += f'{self.left.build(dialect)} * {self.right.build(dialect)}'
            if self.operator == Arithmetic.div:
                ret += f'{self.left.build(dialect)} / {self.right.build(dialect)}'
            ret += ')'
            return ret if self.alias is None else f'{ret} as {self.alias}'
        if dialect == Dialects.POSTGRESQL:
            ret = '('
            if self.operator == Arithmetic.add:
                ret += f'{self.left.build(dialect)} + {self.right.build(dialect)}'
            if self.operator == Arithmetic.sub:
                ret += f'{self.left.build(dialect)} - {self.right.build(dialect)}'
            if self.operator == Arithmetic.mul:
                ret += f'{self.left.build(dialect)} * {self.right.build(dialect)}'
            if self.operator == Arithmetic.div:
                ret += f'{self.left.build(dialect)} / {self.right.build(dialect)}'
            ret += ')'
            return ret if self.alias is None else f'{ret} as {self.alias}'
        if dialect == Dialects.MONGO:
            ret = {}
            if self.operator == Arithmetic.sub:
                ret['$add'] = [self.left.build(dialect), self.right.build(dialect)]
            print(ret)
            return ret


class Column(Term):
    def __init__(self, name, table):
        super(Column, self).__init__(None)

        self._name = name
        self._table = table
        self.alias = None

    def build(self, dialect, depth=0):
        if dialect == Dialects.POSTGRESQL:
            return f'{self._table.build(dialect)}."{self._name}"'


class Table:
    def __init__(self, name, schema=None):
        self._name = name
        self._schema = schema
        self.alias = None

    def __getattr__(self, item):
        return Column(item, self)

    def star(self):
        return Column('*', self)

    def as_(self, alias):
        self.alias = alias
        return self

    def build(self, dialect, depth=0):
        if dialect == Dialects.POSTGRESQL:
            if self._schema is None:
                return f'"{self._name}"'
            else:
                return f'{self._schema.build(dialect)}."{self._name}"'


class Schema:
    def __init__(self, name, database):
        self._name = name
        self._database = database
        self.alias = None

    def __getattr__(self, item):
        return Table(item, self)

    def as_(self, alias):
        self.alias = alias
        return self

    def build(self, dialect, depth=0):
        if dialect == Dialects.POSTGRESQL:
            return f'{self._database.build(dialect)}."{self._name}"'


class Database:
    def __init__(self, name):
        self._name = name
        self.alias = None

    def __getattr__(self, item):
        return Schema(item, self)

    def as_(self, alias):
        self.alias = alias
        return self

    def build(self, dialect, depth=0):
        if dialect == Dialects.POSTGRESQL:
            return f'"{self._name}"'


class Count:
    def __init__(self, column):
        self.column = column
        self.alias = None

    def as_(self, alias):
        self.alias = alias
        return self

    def build(self, dialect, depth=0):
        if dialect == Dialects.MSSQL:
            if isinstance(self.column, Column):
                return f'count({self.column._table._name}.{self.column._name})'
            elif isinstance(self.column, Criterion):
                return f'count({self.column.build(dialect)})'
        if dialect == Dialects.POSTGRESQL:
            if isinstance(self.column, Column):
                return f'count({self.column._table._name}.{self.column._name})'
            elif isinstance(self.column, Criterion):
                return f'count(*) filter(where {self.column.build(dialect)})'
        if dialect == Dialects.MONGO:
            ...


class Array:
    def __init__(self, *column):
        self.column = column
        self.alias = None

        self.dtype = None

    def as_(self, alias):
        self.alias = alias
        return self

    def as_text(self):
        self.dtype = 'text'
        return self

    def build(self, dialect, depth=0):
        if dialect == Dialects.MSSQL:
            raise ValueError
        if dialect == Dialects.POSTGRESQL:
            if isinstance(self.column[0], Column):
                return f'array_agg(({",".join([x._table._name + "." + x._name for x in self.column])}))'
            elif isinstance(self.column[0], Table):
                return f'array_agg({self.column[0]._name})'
            elif isinstance(self.column[0], Query):
                # print('-' * 30)
                # print(self.column[0].build(dialect))
                # print('-' * 30)
                return f'ARRAY({self.column[0].build(dialect)}){f"::{self.dtype}[]" if self.dtype is not None else ""}'
            elif isinstance(self.column[0], Criterion):
                return f'count(*) filter(where {self.column[0].build(dialect)})'
        if dialect == Dialects.MONGO:
            ...


class JSON:
    def __init__(self, **kwargs):
        self.column = kwargs
        self.alias = None

        self.dtype = None
        self.is_array = True

    def as_(self, alias):
        self.alias = alias
        return self

    def as_text(self):
        self.dtype = 'text'
        return self

    def array(self):
        self.is_array = True
        return self

    def object(self):
        self.is_array = False
        return self

    def build(self, dialect, depth=0):
        # print('-' * 30)
        # print(self.column)
        # print('-' * 30)
        if dialect == Dialects.MSSQL:
            raise ValueError
        if dialect == Dialects.POSTGRESQL:
            x = ','.join(["'" + k + "'" + ', ' + v.build(dialect) for k, v in self.column.items()])
            if self.is_array:
                return f"json_agg(json_build_object({x}))"
            else:
                return f"json_build_object({x})"
        if dialect == Dialects.MONGO:
            ...


class Sum:
    def __init__(self, column):
        self.column = column

    def build(self, dialect, depth=0):
        if dialect == Dialects.MSSQL:
            return f'sum({self.column._table._name}.{self.column._name})'
        if dialect == Dialects.POSTGRESQL:
            return f'sum({self.column._table._name}.{self.column._name})'
        if dialect == Dialects.MONGO:
            ...


class Query:
    def __init__(self):
        self.projection = []
        self.condition = None
        self.grouping = []
        self.sort = []
        self.limit = None

        self.tables = []
        self.joins = []

        self.withs = []

        self.alias = None

    def as_table(self):
        return Table(self.alias)

    def as_(self, alias):
        self.alias = alias
        return self

    def with_(self, *query):
        self.withs = query
        return self

    def select(self, *terms):
        self.projection += list(terms)
        return self

    def from_(self, *tables):
        self.tables += list(tables)
        return self

    def join(self, table, term):
        self.joins.append(('inner', table, term))
        return self

    def left_join(self, table, term):
        self.joins.append(('left', table, term))
        return self

    def where(self, term):
        if self.condition is None:
            self.condition = term
        else:
            self.condition = self.condition & term
        return self

    def group(self, *terms):
        self.grouping += list(terms)
        return self

    def order(self, *terms):
        self.sort += list(terms)
        return self

    def top(self, limit):
        self.limit = limit
        return self

    def build(self, dialect, depth=0):
        if dialect == Dialects.POSTGRESQL:
            sql = ''
            if len(self.withs) > 0:
                sql += f'with {",".join([x.alias + " as (" + x.build(dialect) + ")" for x in self.withs])}\n'

            if len(self.projection) == 0:
                raise ValueError('select clause cannot be empty')

            if depth >= 0:
                sql += '('
            sql += 'select ' + ', '.join([x.build(dialect, depth + 1) + " as " + x.alias for x in self.projection]) + '\n'

            if len(self.tables) == 0:
                raise ValueError('from clause cannot be empty')
            sql += 'from ' + ', '.join([x.build(dialect) for x in self.tables]) + '\n'

            if len(self.joins) > 0:
                for join in self.joins:
                    jtype, jtable, condition = join

                    sql += f'{jtype} join {jtable.build(dialect) if not isinstance(jtable, Query) else jtable.as_table().build(dialect)} on {condition.build(dialect)}' + '\n'

            if self.condition is not None:
                sql += 'where ' + self.condition.build(dialect) + '\n'

            if len(self.grouping) > 0:
                sql += 'group by ' + ', '.join([x.build(dialect) for x in self.grouping]) + '\n'

            if len(self.sort) > 0:
                sql += 'order by ' + ', '.join([x.build(dialect) for (x, y) in self.sort]) + '\n'

            if self.limit is not None:
                sql += f'limit {self.limit}\n'

            if depth >= 0:
                sql += ')'

            return sql
        if dialect == Dialects.MONGO:
            # first table will be base table
            select = [x.build(dialect) for x in self.projection]
            where = [x.build(dialect) for x in self.condition]

            ret = []
            for table in self.condition:
                table_criterias = list(filter(lambda x: x.left.table == table, self.condition))
                table_selects = list(filter(lambda x: x.table == table, self.projection))
                ret.append({'$match': table_criterias[0].build(dialect)})
                ret.append({'$project': {table_selects[0].build(dialect): 1}})

            return ret



from __future__ import print_function

import psutil

from joatmon.assistant.task import BaseTask
from joatmon.utility import convert_size


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "system",
            "description": "a function for user to learn current status of the system",
            "parameters": {
                "type": "object",
                "properties": {
                    "action": {
                        "type": "string",
                        "enum": ["memory", "cpu", "disk", "battery"]
                    }
                },
                "required": ["action"]
            }
        }

    def run(self):
        action = self.kwargs.get('action', '') or self.api.input('what do you want the action to be')

        if action == 'memory':
            self.api.output(
                f'Memory Total: {convert_size(psutil.virtual_memory().total)}, '
                f'Memory Used: {convert_size(psutil.virtual_memory().used)}, '
                f'Memory Free: {convert_size(psutil.virtual_memory().free)}, '
                f'Memory Percent: {psutil.virtual_memory().percent}, '
                f'Swap Total: {convert_size(psutil.swap_memory().total)}, '
                f'Swap Used: {convert_size(psutil.swap_memory().used)}, '
                f'Swap Free: {convert_size(psutil.swap_memory().free)}, '
                f'Swap Percent: {psutil.swap_memory().percent}'
            )
        if action == 'cpu':
            self.api.output(
                f'CPU Total Usage: {psutil.cpu_percent(percpu=False)}, '
                f'CPU Per Usage: {psutil.cpu_percent(percpu=True)}, '
                f'CPU Count: {psutil.cpu_count(logical=False)}, '
                f'CPU Count Logical: {psutil.cpu_count(logical=True)}'
            )
        if action == 'disk':
            for d in psutil.disk_partitions():
                self.api.output(
                    f'Disk Device: {d.device}, '
                    f'Disk Mount: {d.mountpoint}, '
                    f'Disk File System Type: {d.fstype}'
                )

                self.api.output(
                    f'Disk Total: {convert_size(psutil.disk_usage(d.device).total)}, '
                    f'Disk Used: {convert_size(psutil.disk_usage(d.device).used)}, '
                    f'Disk Free: {convert_size(psutil.disk_usage(d.device).free)}, '
                    f'Disk Percent: {psutil.disk_usage(d.device).percent}'
                )

        if action == 'battery':
            self.api.output(
                f'Battery Percent: {psutil.sensors_battery().percent}, '
                f'Battery Plugged: {psutil.sensors_battery().power_plugged}, '
                f'Batter Left: {psutil.sensors_battery().secsleft}'
            )
        # if action == 'process':
        #     for p in psutil.pids():
        #         process = psutil.Process(p)
        #         print(process)

        if not self.stop_event.is_set():
            self.stop_event.set()



import functools
import inspect

from joatmon import context
from joatmon.utility import (
    to_case,
    to_enumerable
)


def get(func):
    func.__method__ = 'get'

    return func


def post(func):
    func.__method__ = 'post'

    return func


def incoming(case, json, arg, form):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):  # if func.method is get need to do with args, else with json
            c = context.get_value(case).get()
            c = 'snake'

            if (j := context.get_value(json).get()) is not None:
                if c is not None:
                    kwargs.update(to_case(c, j))
                else:
                    kwargs.update(j)

            if (a := context.get_value(arg).get()) is not None:
                if c is not None:
                    kwargs.update(to_case(c, a))
                else:
                    kwargs.update(a)

            if (f := context.get_value(form).get()) is not None:
                if c is not None:
                    kwargs.update(to_case(c, f))
                else:
                    kwargs.update(f)

            return await func(*args, **kwargs)

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator


def wrap(func):
    @functools.wraps(func)
    async def _wrapper(*args, **kwargs):
        try:
            data = await func(*args, **kwargs)
            return {'data': to_enumerable(data), 'error': None, 'success': True}
        except Exception as ex:
            return {'data': None, 'error': str(ex), 'success': False}

    _wrapper.__signature__ = inspect.signature(func)
    return _wrapper


def outgoing(case):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            response = await func(*args, **kwargs)
            c = context.get_value(case).get()
            c = 'pascal'
            if c is not None:
                response = to_case(c, response)
            return response

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator


def ip_limit(interval, cache, ip):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            c = context.get_value(cache)
            ip_value = context.get_value(ip).get()

            key = f'limit_request:{func.__qualname__}:{ip_value}'

            if await c.get(key) is not None:
                raise ValueError('too_many_request')
            await c.add(key, 1, duration=interval)

            response = await func(*args, **kwargs)
            return response

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator


def limit(interval, cache):
    def _decorator(func):
        @functools.wraps(func)
        async def _wrapper(*args, **kwargs):
            c = context.get_value(cache)

            key = f'limit_request:{func.__qualname__}'

            if await c.get(key) is not None:
                raise ValueError('too_many_request')

            await c.add(key, 1, duration=interval)

            response = await func(*args, **kwargs)
            return response

        _wrapper.__signature__ = inspect.signature(func)
        return _wrapper

    return _decorator



import asyncio
import logging
import re
from asyncio import CancelledError
from collections import deque
from functools import partial

import websockets
from transitions import (
    Machine,
    MachineError
)

from .adapters import (
    ReaderAdapter,
    StreamReaderAdapter,
    StreamWriterAdapter,
    WebSocketsReader,
    WebSocketsWriter,
    WriterAdapter
)
from .errors import (
    HZMQTTException,
    MQTTException
)
from .mqtt.handler import BrokerProtocolHandler
from .session import Session
from .utils import (
    format_client_message,
    gen_client_id
)

_defaults = {
    'timeout-disconnect-delay': 2,
    'auth': {
        'allow-anonymous': True,
        'password-file': None
    },
}

EVENT_BROKER_PRE_START = 'broker_pre_start'
EVENT_BROKER_POST_START = 'broker_post_start'
EVENT_BROKER_PRE_SHUTDOWN = 'broker_pre_shutdown'
EVENT_BROKER_POST_SHUTDOWN = 'broker_post_shutdown'
EVENT_BROKER_CLIENT_CONNECTED = 'broker_client_connected'
EVENT_BROKER_CLIENT_DISCONNECTED = 'broker_client_disconnected'
EVENT_BROKER_CLIENT_SUBSCRIBED = 'broker_client_subscribed'
EVENT_BROKER_CLIENT_UNSUBSCRIBED = 'broker_client_unsubscribed'
EVENT_BROKER_MESSAGE_RECEIVED = 'broker_message_received'


class BrokerException(BaseException):
    pass


class RetainedApplicationMessage:
    __slots__ = ('source_session', 'topic', 'data', 'qos')

    def __init__(self, source_session, topic, data, qos=None):
        self.source_session = source_session
        self.topic = topic
        self.data = data
        self.qos = qos


class Server:
    def __init__(self, listener_name, server_instance, max_connections=-1, loop=None):
        self.logger = logging.getLogger(__name__)
        self.instance = server_instance
        self.conn_count = 0
        self.listener_name = listener_name
        if loop is not None:
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()

        self.max_connections = max_connections
        if self.max_connections > 0:
            self.semaphore = asyncio.Semaphore(self.max_connections, loop=self._loop)
        else:
            self.semaphore = None

    async def acquire_connection(self):
        if self.semaphore:
            await self.semaphore.acquire()
        self.conn_count += 1
        if self.max_connections > 0:
            self.logger.info("Listener '%s': %d/%d connections acquired" % (self.listener_name, self.conn_count, self.max_connections))
        else:
            self.logger.info("Listener '%s': %d connections acquired" % (self.listener_name, self.conn_count))

    def release_connection(self):
        if self.semaphore:
            self.semaphore.release()
        self.conn_count -= 1
        if self.max_connections > 0:
            self.logger.info("Listener '%s': %d/%d connections acquired" % (self.listener_name, self.conn_count, self.max_connections))
        else:
            self.logger.info("Listener '%s': %d connections acquired" % (self.listener_name, self.conn_count))

    async def close_instance(self):
        if self.instance:
            self.instance.close()
            await self.instance.wait_closed()


class Broker:
    states = ['new', 'starting', 'started', 'not_started', 'stopping', 'stopped', 'not_stopped', 'stopped']

    def __init__(self, config=None, loop=None):
        self.logger = logging.getLogger(__name__)
        self.config = _defaults
        if config is not None:
            self.config.update(config)
        self._build_listeners_config(self.config)

        if loop is not None:
            self._loop = loop
        else:
            self._loop = asyncio.get_event_loop()

        self._servers = dict()
        self._init_states()
        self._sessions = dict()
        self._subscriptions = dict()
        self._retained_messages = dict()
        self._broadcast_queue = asyncio.Queue(loop=self._loop)

        self._broadcast_task = None

    def _build_listeners_config(self, broker_config):
        self.listeners_config = dict()
        try:
            listeners_config = broker_config['listeners']
            defaults = listeners_config['default']
            for listener in listeners_config:
                config = dict(defaults)
                config.update(listeners_config[listener])
                self.listeners_config[listener] = config
        except KeyError as ke:
            raise BrokerException("Listener config not found invalid: %s" % ke)

    def _init_states(self):
        self.transitions = Machine(states=Broker.states, initial='new')
        self.transitions.add_transition(trigger='start', source='new', dest='starting')
        self.transitions.add_transition(trigger='starting_fail', source='starting', dest='not_started')
        self.transitions.add_transition(trigger='starting_success', source='starting', dest='started')
        self.transitions.add_transition(trigger='shutdown', source='started', dest='stopping')
        self.transitions.add_transition(trigger='stopping_success', source='stopping', dest='stopped')
        self.transitions.add_transition(trigger='stopping_failure', source='stopping', dest='not_stopped')
        self.transitions.add_transition(trigger='start', source='stopped', dest='starting')

    async def start(self):
        try:
            self._sessions = dict()
            self._subscriptions = dict()
            self._retained_messages = dict()
            self.transitions.start()
            self.logger.debug("Broker starting")
        except (MachineError, ValueError) as exc:
            self.logger.warning("[WARN-0001] Invalid method call at this moment: %s" % exc)
            raise BrokerException("Broker instance can't be started: %s" % exc)

        self.logger.info('EVENT_BROKER_PRE_START')
        try:
            for listener_name in self.listeners_config:
                listener = self.listeners_config[listener_name]

                if 'bind' not in listener:
                    self.logger.debug("Listener configuration '%s' is not bound" % listener_name)
                else:
                    try:
                        max_connections = listener['max_connections']
                    except KeyError:
                        max_connections = -1

                    address, s_port = listener['bind'].split(':')

                    try:
                        port = int(s_port)
                    except ValueError:
                        raise BrokerException("Invalid port value in bind value: %s" % listener['bind'])

                    if listener['type'] == 'tcp':
                        cb_partial = partial(self.stream_connected, listener_name=listener_name)
                        instance = await asyncio.start_server(cb_partial, address, port, reuse_address=True, loop=self._loop)
                        self._servers[listener_name] = Server(listener_name, instance, max_connections, self._loop)
                    elif listener['type'] == 'ws':
                        cb_partial = partial(self.ws_connected, listener_name=listener_name)
                        instance = await websockets.serve(cb_partial, address, port, loop=self._loop, subprotocols=['mqtt'])
                        self._servers[listener_name] = Server(listener_name, instance, max_connections, self._loop)

                    self.logger.info("Listener '%s' bind to %s (max_connections=%d)" % (listener_name, listener['bind'], max_connections))

            self.transitions.starting_success()
            self.logger.info('EVENT_BROKER_POST_START')

            self._broadcast_task = asyncio.ensure_future(self._broadcast_loop(), loop=self._loop)

            self.logger.debug("Broker started")
        except Exception as e:
            self.logger.error("Broker startup failed: %s" % e)
            self.transitions.starting_fail()
            raise BrokerException("Broker instance can't be started: %s" % e)

    async def shutdown(self):
        try:
            self._sessions = dict()
            self._subscriptions = dict()
            self._retained_messages = dict()
            self.transitions.shutdown()
        except (MachineError, ValueError) as exc:
            self.logger.debug("Invalid method call at this moment: %s" % exc)
            raise BrokerException("Broker instance can't be stopped: %s" % exc)

        self.logger.info('EVENT_BROKER_PRE_SHUTDOWN')

        if self._broadcast_task:
            self._broadcast_task.cancel()
        if self._broadcast_queue.qsize() > 0:
            self.logger.warning("%d messages not broadcasted" % self._broadcast_queue.qsize())

        for listener_name in self._servers:
            server = self._servers[listener_name]
            await server.close_instance()
        self.logger.debug("Broker closing")
        self.logger.info("Broker closed")
        self.logger.info('EVENT_BROKER_POST_SHUTDOWN')
        self.transitions.stopping_success()

    async def internal_message_broadcast(self, topic, data, qos=None):
        return await self._broadcast_message(None, topic, data)

    async def ws_connected(self, websocket, uri, listener_name):
        await self.client_connected(listener_name, WebSocketsReader(websocket), WebSocketsWriter(websocket))

    async def stream_connected(self, reader, writer, listener_name):
        await self.client_connected(listener_name, StreamReaderAdapter(reader), StreamWriterAdapter(writer))

    async def client_connected(self, listener_name, reader: ReaderAdapter, writer: WriterAdapter):
        print('new client connected')
        server = self._servers.get(listener_name, None)
        if not server:
            raise BrokerException("Invalid listener name '%s'" % listener_name)
        await server.acquire_connection()

        remote_address, remote_port = writer.get_peer_info()
        self.logger.info("Connection from %s:%d on listener '%s'" % (remote_address, remote_port, listener_name))

        try:
            handler, client_session = await BrokerProtocolHandler.init_from_connect(reader, writer, loop=self._loop)
        except HZMQTTException as exc:
            self.logger.warning("[MQTT-3.1.0-1] %s: Can't read first packet an CONNECT: %s" % (format_client_message(address=remote_address, port=remote_port), exc))
            self.logger.debug("Connection closed")
            return
        except MQTTException as me:
            self.logger.error('Invalid connection from %s : %s' % (format_client_message(address=remote_address, port=remote_port), me))
            await writer.close()
            self.logger.debug("Connection closed")
            return

        if client_session.clean_session:
            if client_session.client_id is not None and client_session.client_id != "":
                self.delete_session(client_session.client_id)
            else:
                client_session.client_id = gen_client_id()
            client_session.parent = 0
        else:
            if client_session.client_id in self._sessions:
                self.logger.debug("Found old session %s" % repr(self._sessions[client_session.client_id]))
                (client_session, h) = self._sessions[client_session.client_id]
                client_session.parent = 1
            else:
                client_session.parent = 0
        if client_session.keep_alive > 0:
            client_session.keep_alive += self.config['timeout-disconnect-delay']
        self.logger.debug("Keep-alive timeout=%d" % client_session.keep_alive)

        handler.attach(client_session, reader, writer)
        self._sessions[client_session.client_id] = (client_session, handler)

        authenticated = await self.authenticate(client_session, self.listeners_config[listener_name])
        if not authenticated:
            await writer.close()
            server.release_connection()
            return

        while True:
            try:
                client_session.transitions.connect()
                break
            except (MachineError, ValueError):
                self.logger.warning("Client %s is reconnecting too quickly, make it wait" % client_session.client_id)
                await asyncio.sleep(1, loop=self._loop)
        await handler.mqtt_connack_authorize(authenticated)

        self.logger.info(f'EVENT_BROKER_CLIENT_CONNECTED: {client_session.client_id}')

        self.logger.debug("%s Start messages handling" % client_session.client_id)
        await handler.start()
        self.logger.debug("Retained messages queue size: %d" % client_session.retained_messages.qsize())
        await self.publish_session_retained_messages(client_session)

        disconnect_waiter = asyncio.ensure_future(handler.wait_disconnect(), loop=self._loop)
        subscribe_waiter = asyncio.ensure_future(handler.get_next_pending_subscription(), loop=self._loop)
        unsubscribe_waiter = asyncio.ensure_future(handler.get_next_pending_unsubscription(), loop=self._loop)
        wait_deliver = asyncio.ensure_future(handler.mqtt_deliver_next_message(), loop=self._loop)
        connected = True
        while connected:
            try:
                done, pending = await asyncio.wait(
                    [disconnect_waiter, subscribe_waiter, unsubscribe_waiter, wait_deliver],
                    return_when=asyncio.FIRST_COMPLETED, loop=self._loop
                )
                if disconnect_waiter in done:
                    print('disconnection')
                    result = disconnect_waiter.result()
                    self.logger.debug("%s Result from wait_diconnect: %s" % (client_session.client_id, result))
                    if result is None:
                        self.logger.debug("Will flag: %s" % client_session.will_flag)
                        if client_session.will_flag:
                            self.logger.debug("Client %s disconnected abnormally, sending will message" % format_client_message(client_session))
                            await self._broadcast_message(client_session, client_session.will_topic, client_session.will_message, client_session.will_qos)
                            if client_session.will_retain:
                                self.retain_message(client_session, client_session.will_topic, client_session.will_message, client_session.will_qos)
                    self.logger.debug("%s Disconnecting session" % client_session.client_id)
                    await self._stop_handler(handler)
                    client_session.transitions.disconnect()
                    self.logger.info(f'EVENT_BROKER_CLIENT_DISCONNECTED: {client_session.client_id}')
                    connected = False
                if unsubscribe_waiter in done:
                    print('unsubscribed')
                    self.logger.debug("%s handling unsubscription" % client_session.client_id)
                    unsubscription = unsubscribe_waiter.result()
                    for topic in unsubscription['topics']:
                        self._del_subscription(topic, client_session)
                        self.logger.info(f'EVENT_BROKER_CLIENT_UNSUBSCRIBED: {client_session.client_id}, {topic}')
                    await handler.mqtt_acknowledge_unsubscription(unsubscription['packet_id'])
                    unsubscribe_waiter = asyncio.Task(handler.get_next_pending_unsubscription(), loop=self._loop)
                if subscribe_waiter in done:
                    print('subscribed')
                    self.logger.debug("%s handling subscription" % client_session.client_id)
                    subscriptions = subscribe_waiter.result()
                    return_codes = []
                    for subscription in subscriptions['topics']:
                        result = await self.add_subscription(subscription, client_session)
                        return_codes.append(result)
                    await handler.mqtt_acknowledge_subscription(subscriptions['packet_id'], return_codes)
                    for index, subscription in enumerate(subscriptions['topics']):
                        if return_codes[index] != 0x80:
                            self.logger.info(f'EVENT_BROKER_CLIENT_SUBSCRIBED: {client_session.client_id}, {subscription[0]}, {subscription[1]}')
                            await self.publish_retained_messages_for_subscription(subscription, client_session)
                    subscribe_waiter = asyncio.Task(handler.get_next_pending_subscription(), loop=self._loop)
                    self.logger.debug(repr(self._subscriptions))
                if wait_deliver in done:
                    print('delivered')
                    if self.logger.isEnabledFor(logging.DEBUG):
                        self.logger.debug("%s handling message delivery" % client_session.client_id)
                    app_message = wait_deliver.result()
                    if not app_message.topic:
                        self.logger.warning("[MQTT-4.7.3-1] - %s invalid TOPIC sent in PUBLISH message, closing connection" % client_session.client_id)
                        break
                    if "#" in app_message.topic or "+" in app_message.topic:
                        self.logger.warning("[MQTT-3.3.2-2] - %s invalid TOPIC sent in PUBLISH message, closing connection" % client_session.client_id)
                        break

                    self.logger.info(f'EVENT_BROKER_MESSAGE_RECEIVED: {client_session.client_id}, {app_message}')
                    await self._broadcast_message(client_session, app_message.topic, app_message.data)
                    if app_message.publish_packet.retain_flag:
                        self.retain_message(client_session, app_message.topic, app_message.data, app_message.qos)
                    wait_deliver = asyncio.Task(handler.mqtt_deliver_next_message(), loop=self._loop)
            except asyncio.CancelledError:
                self.logger.debug("Client loop cancelled")
                break
        disconnect_waiter.cancel()
        subscribe_waiter.cancel()
        unsubscribe_waiter.cancel()
        wait_deliver.cancel()

        self.logger.debug("%s Client disconnected" % client_session.client_id)
        server.release_connection()

    def _init_handler(self, session, reader, writer):
        handler = BrokerProtocolHandler(loop=self._loop)
        handler.attach(session, reader, writer)
        return handler

    async def _stop_handler(self, handler):
        try:
            await handler.stop()
        except Exception as e:
            self.logger.error(e)

    async def authenticate(self, session: Session, listener):
        auth_result = True
        return auth_result

    async def topic_filtering(self, session: Session, topic):
        topic_result = True
        return topic_result

    def retain_message(self, source_session, topic_name, data, qos=None):
        if data is not None and data != b'':
            self.logger.debug("Retaining message on topic %s" % topic_name)
            retained_message = RetainedApplicationMessage(source_session, topic_name, data, qos)
            self._retained_messages[topic_name] = retained_message
        else:
            if topic_name in self._retained_messages:
                self.logger.debug("Clear retained messages for topic '%s'" % topic_name)
                del self._retained_messages[topic_name]

    async def add_subscription(self, subscription, session):
        try:
            a_filter = subscription[0]
            if '#' in a_filter and not a_filter.endswith('#'):
                return 0x80
            if a_filter != "+":
                if '+' in a_filter:
                    if "/+" not in a_filter and "+/" not in a_filter:
                        return 0x80
            permitted = await self.topic_filtering(session, topic=a_filter)
            if not permitted:
                return 0x80
            qos = subscription[1]
            if 'max-qos' in self.config and qos > self.config['max-qos']:
                qos = self.config['max-qos']
            if a_filter not in self._subscriptions:
                self._subscriptions[a_filter] = []
            already_subscribed = next(
                (s for (s, qos) in self._subscriptions[a_filter] if s.client_id == session.client_id), None
            )
            if not already_subscribed:
                self._subscriptions[a_filter].append((session, qos))
            else:
                self.logger.debug("Client %s has already subscribed to %s" % (format_client_message(session=session), a_filter))
            return qos
        except KeyError:
            return 0x80

    def _del_subscription(self, a_filter, session):
        deleted = 0
        try:
            subscriptions = self._subscriptions[a_filter]
            for index, (sub_session, qos) in enumerate(subscriptions):
                if sub_session.client_id == session.client_id:
                    self.logger.debug("Removing subscription on topic '%s' for client %s" % (a_filter, format_client_message(session=session)))
                    subscriptions.pop(index)
                    deleted += 1
                    break
        except KeyError:
            pass
        finally:
            return deleted

    def _del_all_subscriptions(self, session):
        filter_queue = deque()
        for topic in self._subscriptions:
            if self._del_subscription(topic, session):
                filter_queue.append(topic)
        for topic in filter_queue:
            if not self._subscriptions[topic]:
                del self._subscriptions[topic]

    @staticmethod
    def matches(topic, a_filter):
        if "#" not in a_filter and "+" not in a_filter:
            return a_filter == topic
        else:
            match_pattern = re.compile(a_filter.replace('#', '.*').replace('$', '\$').replace('+', '[/\$\s\w\d]+'))
            return match_pattern.match(topic)

    async def _broadcast_loop(self):
        running_tasks = deque()
        try:
            while True:
                while running_tasks and running_tasks[0].done():
                    task = running_tasks.popleft()
                    try:
                        task.result()
                    except Exception:
                        pass
                broadcast = await self._broadcast_queue.get()
                if self.logger.isEnabledFor(logging.DEBUG):
                    self.logger.debug("broadcasting %r" % broadcast)
                for k_filter in self._subscriptions:
                    print(k_filter, broadcast['topic'])
                    if broadcast['topic'].startswith("$") and (k_filter.startswith("+") or k_filter.startswith("#")):
                        self.logger.debug("[MQTT-4.7.2-1] - ignoring brodcasting $ topic to subscriptions starting with + or #")
                    elif self.matches(broadcast['topic'], k_filter):
                        subscriptions = self._subscriptions[k_filter]
                        for (target_session, qos) in subscriptions:
                            if 'qos' in broadcast:
                                qos = broadcast['qos']
                            if target_session.transitions.state == 'connected':
                                if self.logger.isEnabledFor(logging.DEBUG):
                                    self.logger.debug(
                                        "broadcasting application message from %s on topic '%s' to %s" % (
                                            format_client_message(session=broadcast['session']), broadcast['topic'], format_client_message(session=target_session))
                                    )
                                handler = self._get_handler(target_session)
                                task = asyncio.ensure_future(
                                    handler.mqtt_publish(broadcast['topic'], broadcast['data'], qos, retain=False),
                                    loop=self._loop
                                )
                                running_tasks.append(task)
                            elif qos is not None and qos > 0:
                                if self.logger.isEnabledFor(logging.DEBUG):
                                    self.logger.debug(
                                        "retaining application message from %s on topic '%s' to client '%s'" % (
                                            format_client_message(session=broadcast['session']), broadcast['topic'], format_client_message(session=target_session))
                                    )
                                retained_message = RetainedApplicationMessage(
                                    broadcast['session'], broadcast['topic'], broadcast['data'], qos
                                )
                                await target_session.retained_messages.put(retained_message)
                                if self.logger.isEnabledFor(logging.DEBUG):
                                    self.logger.debug(f'target_session.retained_messages={target_session.retained_messages.qsize()}')
        except CancelledError:
            if running_tasks:
                await asyncio.wait(running_tasks, loop=self._loop)
            raise

    async def _broadcast_message(self, session, topic, data, force_qos=None):
        broadcast = {'session': session, 'topic': topic, 'data': data}
        if force_qos:
            broadcast['qos'] = force_qos
        await self._broadcast_queue.put(broadcast)

    async def publish_session_retained_messages(self, session):
        self.logger.debug("Publishing %d messages retained for session %s" % (session.retained_messages.qsize(), format_client_message(session=session)))
        publish_tasks = []
        handler = self._get_handler(session)
        while not session.retained_messages.empty():
            retained = await session.retained_messages.get()
            publish_tasks.append(
                asyncio.ensure_future(
                    handler.mqtt_publish(retained.topic, retained.data, retained.qos, True), loop=self._loop
                )
            )
        if publish_tasks:
            await asyncio.wait(publish_tasks, loop=self._loop)

    async def publish_retained_messages_for_subscription(self, subscription, session):
        self.logger.debug("Begin broadcasting messages retained due to subscription on '%s' from %s" % (subscription[0], format_client_message(session=session)))
        publish_tasks = []
        handler = self._get_handler(session)
        for d_topic in self._retained_messages:
            self.logger.debug("matching : %s %s" % (d_topic, subscription[0]))
            if self.matches(d_topic, subscription[0]):
                self.logger.debug("%s and %s match" % (d_topic, subscription[0]))
                retained = self._retained_messages[d_topic]
                publish_tasks.append(asyncio.Task(handler.mqtt_publish(retained.topic, retained.data, subscription[1], True), loop=self._loop))
        if publish_tasks:
            await asyncio.wait(publish_tasks, loop=self._loop)
        self.logger.debug("End broadcasting messages retained due to subscription on '%s' from %s" % (subscription[0], format_client_message(session=session)))

    def delete_session(self, client_id):
        try:
            session = self._sessions[client_id][0]
        except KeyError:
            session = None
        if session is None:
            self.logger.debug("Delete session : session %s doesn't exist" % client_id)
            return

        self.logger.debug("deleting session %s subscriptions" % repr(session))
        self._del_all_subscriptions(session)

        self.logger.debug("deleting existing session %s" % repr(self._sessions[client_id]))
        del self._sessions[client_id]

    def _get_handler(self, session):
        client_id = session.client_id
        if client_id:
            try:
                return self._sessions[client_id][1]
            except KeyError:
                pass
        return None



import numpy as np

from joatmon.nn.core import Loss

__all__ = ['MSELoss']


class MSELoss(Loss):
    def __init__(self):
        super(MSELoss, self).__init__()

        self._loss = None

    def forward(self, prediction, target) -> np.array:
        self._loss = (((prediction - target) ** 2) / np.prod(target.shape)).summation()
        return self._loss



def _check_input(data):
    def _helper(_data, _length=None, _dtype=None):
        data_length = None
        data_type = type(_data)

        if isinstance(_data, (tuple, list)):
            data_length = len(_data)
            if len(_data) == 0:
                raise ValueError('empty list cannot be used')

            if _length is not None and len(_data) != _length:
                raise ValueError(f'shape mismatch')

            _check_input(_data)
        elif not isinstance(_data, (bool, int, float)):
            raise ValueError('data has to be one of the following types: (bool, int, float)')

        if _dtype is not None and type(_data) != _dtype:
            raise ValueError(f'data type mismatch between {type(_data)} and {_dtype}')

        return data_length, data_type

    if not isinstance(data, (tuple, list)):
        raise ValueError('input data has to be either tuple or list type')

    length = None
    dtype = None
    for _data in data:
        length, dtype = _helper(_data, length, dtype)


def _normalize_input(data):
    return data


class Array:
    def __init__(self, data=None):
        _check_input(data)
        _normalize_input(data)

        self._data = data
        self._device = 'cpu'
        self._dtype = 'float32'

    def __repr__(self):
        return str(self)

    def __str__(self):
        return f'Array(data={self.data}, dtype={self.dtype}, shape={self.shape})'

    def __bool__(self):
        return True

    def __copy__(self):
        return self.copy()

    def __deepcopy__(self, memodict=None):
        return self.copy()

    def __iter__(self):
        for data in self.data:
            yield data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        return self.getitem(idx)

    def __setitem__(self, key, value):
        self.setitem(key, value)

    def __abs__(self):
        return self.abs()

    def __neg__(self):
        return self.negative()

    def __round__(self, n=None):
        pass

    def __floor__(self):
        pass

    def __ceil__(self):
        pass

    def __add__(self, other):
        return self.add(other)

    def __iadd__(self, other):
        return self.add(other, out=self)

    def __sub__(self, other):
        return self.sub(other)

    def __isub__(self, other):
        return self.sub(other, out=self)

    def __mul__(self, other):
        return self.mul(other)

    def __imul__(self, other):
        return self.mul(other, out=self)

    def __truediv__(self, other):
        return self.truediv(other)

    def __itruediv__(self, other):
        return self.truediv(other, out=self)

    def __floordiv__(self, other):
        return self.floordiv(other)

    def __ifloordiv__(self, other):
        return self.floordiv(other, out=self)

    def __pow__(self, power, modulo=None):
        return self.power(power)

    def __ipow__(self, other):
        return self.power(other, out=self)

    def __lt__(self, other):
        return self.lt(other)

    def __le__(self, other):
        return self.le(other)

    def __gt__(self, other):
        return self.gt(other)

    def __ge__(self, other):
        return self.ge(other)

    def __eq__(self, other):
        return self.eq(other)

    def __ne__(self, other):
        return self.ne(other)

    def astype(self, dtype):
        ...

    def copy(self):
        ...

    def repeat(self):
        ...

    def split(self):
        ...

    def tolist(self):
        ...

    def getitem(self, idx):
        ...

    def take_along_axis(self):
        ...

    def setitem(self, idx, value):
        ...

    def put_along_axis(self):
        ...

    def where(self):
        ...

    def indices(self):
        ...

    def dim(self):
        ...

    def size(self, dim=None):
        ...

    def flatten(self):
        ...

    def reshape(self, size):
        ...

    def squeeze(self):
        ...

    def expand_dims(self):
        ...

    def pad(self):
        ...

    def transpose(self):
        ...

    def fill(self, value):
        ...

    def abs(self):
        ...

    def negative(self, *, out=None):
        ...

    def round(self, *, out=None):
        ...

    def floor(self, *, out=None):
        ...

    def ceil(self, *, out=None):
        ...

    def sqrt(self, *, out=None):
        ...

    def square(self, *, out=None):
        ...

    def clip(self, min_value, max_value, *, out=None):
        ...

    def exp(self, *, out=None):
        ...

    def tanh(self, *, out=None):
        ...

    def sum(self):
        ...

    def mean(self):
        ...

    def median(self):
        ...

    def var(self):
        ...

    def std(self):
        ...

    def prod(self):
        ...

    def unique(self):
        ...

    def argmax(self):
        ...

    def argmin(self):
        ...

    def amax(self):
        ...

    def amin(self):
        ...

    def add(self, other, *, out=None):
        ...

    def sub(self, other, *, out=None):
        ...

    def mul(self, other, *, out=None):
        ...

    def truediv(self, other, *, out=None):
        ...

    def floordiv(self, other, *, out=None):
        ...

    def power(self, p, *, out=None):
        ...

    def lt(self, other):
        ...

    def le(self, other):
        ...

    def gt(self, other):
        ...

    def ge(self, other):
        ...

    def eq(self, other):
        ...

    def ne(self, other):
        ...

    @property
    def device(self):
        return self._device

    @property
    def data(self):
        return self._data

    # noinspection PyPep8Naming
    @property
    def T(self):
        return self.transpose()

    @property
    def ndim(self):
        return self.dim()

    @property
    def shape(self):
        return self.size()

    @property
    def dtype(self):
        return self._dtype



import binascii
import re
import struct
import traceback

from joatmon.system.memory.address import Address
from joatmon.system.memory.core import type_unpack
from joatmon.system.memory.process import (
    Process,
    ProcessException
)

REGEX_TYPE = type(re.compile("^plop$"))

PAGE_READONLY = 2
PAGE_READWRITE = 4


class Worker(object):
    def __init__(self, pid=None, name=None, debug=True):
        self.process = Process(name=name, pid=pid, debug=debug)

    def __enter__(self):
        return self

    def __exit__(self):
        self.process.close()

    def address(self, value, default_type='uint'):
        return Address(value, process=self.process, default_type=default_type)

    def memory_replace_unicode(self, regex, replace):
        regex = regex.encode('utf-16-le')
        replace = replace.encode('utf-16-le')
        return self.memory_replace(re.compile(regex, re.UNICODE), replace)

    def memory_replace(self, regex, replace):
        succeed = True
        for _, start_offset in self.memory_search(regex, of_type='re'):
            if self.process.write_bytes(start_offset, replace) == 1:
                print('Write at offset %s succeeded !' % start_offset)
            else:
                succeed = False
                print('Write at offset %s failed !' % start_offset)

        return succeed

    def memory_search_unicode(self, regex):
        regex = regex.encode('utf-16-le')
        for _name, _address in self.memory_search(regex, of_type='re'):
            yield _name, _address

    def group_search(self, group, start_offset=None, end_offset=None):
        regex = ''
        for value, of_type in group:
            if of_type == 'f' or of_type == 'float':
                float_value = struct.pack('<f', float(value))
                regex += '..' + float_value[2:4]
            else:
                raise NotImplementedError('unknown type %s' % of_type)

        return self.memory_search(regex, of_type='re', start_offset=start_offset, end_offset=end_offset)

    def search_address(self, search):
        _address = '%08X' % search
        print('searching address %s' % _address)
        regex = ''
        for i in range(len(_address) - 2, -1, -2):
            regex += binascii.unhexlify(_address[i:i + 2])

        for _, _address in self.memory_search(re.escape(regex), of_type='re'):
            yield _address

    def parse_re_function(self, byte_array, value, offset):
        for name, regex in value:
            for res in regex.finditer(byte_array):
                yield name, self.address(offset + res.start(), 'bytes')

    def parse_float_function(self, byte_array, value, offset):
        for index in range(0, len(byte_array)):
            try:
                structure_type, structure_len = type_unpack('float')
                temp_val = struct.unpack(structure_type, byte_array[index:index + 4])[0]
                if int(value) == int(temp_val):
                    s_offset = offset + index
                    yield self.address(s_offset, 'float')
            except Exception as ex:
                print(str(ex))

    @staticmethod
    def parse_named_groups_function(byte_array, value, _):
        for name, regex in value:
            for res in regex.finditer(byte_array):
                yield name, res.groupdict()

    @staticmethod
    def parse_groups_function(byte_array, value, _):
        for name, regex in value:
            for res in regex.finditer(byte_array):
                yield name, res.groups()

    def parse_any_function(self, byte_array, value, offset):
        index = byte_array.find(value)
        while index != -1:
            soffset = offset + index
            yield self.address(soffset, 'bytes')
            index = byte_array.find(value, index + 1)

    def memory_search(self, value, of_type='match', protect=PAGE_READWRITE | PAGE_READONLY, optimizations=None, start_offset=None, end_offset=None):
        if of_type == 're' or of_type == 'groups' or of_type == 'ngroups':
            if type(value) is not list:
                value = [value]

            temp = []
            for reg in value:
                if type(reg) is tuple:
                    name = reg[0]
                    if type(reg[1]) != REGEX_TYPE:
                        regex = re.compile(reg[1], re.IGNORECASE)
                    else:
                        regex = reg[1]
                elif type(reg) == REGEX_TYPE:
                    name = ''
                    regex = reg
                else:
                    name = ''
                    regex = re.compile(reg, re.IGNORECASE)

                temp.append((name, regex))
            value = temp
        elif (of_type != 'match' and
              of_type != 'group' and
              of_type != 're' and
              of_type != 'groups' and
              of_type != 'ngroups' and
              of_type != 'lambda'):
            structure_type, structure_len = type_unpack(of_type)
            value = struct.pack(structure_type, value)

        if of_type == 're':
            parser = self.parse_re_function
        elif of_type == 'groups':
            parser = self.parse_groups_function
        elif of_type == 'ngroups':
            parser = self.parse_named_groups_function
        elif of_type == 'float':
            parser = self.parse_float_function
        elif of_type == 'lambda':
            parser = value
        else:
            parser = self.parse_any_function

        if not self.process.is_process_open:
            raise ProcessException("Can't read_bytes, process %s is not open" % self.process.pid)

        for offset, chunk_size in self.process.iter_region(
                start_offset=start_offset,
                end_offset=end_offset,
                protect=protect,
                optimizations=optimizations
        ):
            bytes_array = b''
            current_offset = offset
            chunk_read = 0
            chunk_exc = False
            while chunk_read < chunk_size:
                try:
                    bytes_array += self.process.read_bytes(current_offset, chunk_size)
                except IOError as e:
                    print()
                    traceback.format_exc()
                    if e.errno == 13:
                        raise
                    else:
                        print(e)
                    chunk_exc = True
                    break
                except Exception as ex:
                    print(str(ex))
                    chunk_exc = True
                    break
                finally:
                    current_offset += chunk_size
                    chunk_read += chunk_size

            if chunk_exc:
                continue

            if bytes_array:
                if of_type == "lambda":
                    for result in parser(bytes_array, offset):
                        yield result
                else:
                    for result in parser(bytes_array, value, offset):
                        yield result



import gym


class CoreSpace(gym.Space):
    def __init__(self, shape=None, dtype=None):
        super(CoreSpace, self).__init__(shape, dtype)

    @property
    def is_np_flattenable(self):
        raise NotImplementedError

    def sample(self, mask=None):
        raise NotImplementedError

    def contains(self, x) -> bool:
        raise NotImplementedError


class CoreEnv(gym.Env):
    """
    The abstract game class that is used by all agents. This class has the exact same API that OpenAI Gym uses so that integrating
    with it is trivial. In contrast to the OpenAI Gym implementation, this class only defines the abstract methods without any actual implementation.

    To implement your own game, you need to define the following methods:

    - `seed`
    - `reset`
    - `step`
    - `render`
    - `close`

    Refer to the [Gym documentation](https://gym.openai.com/docs/#environment).
    """

    reward_range = (-float("inf"), float("inf"))
    action_space: CoreSpace
    observation_space: CoreSpace

    def __init__(self):
        super(CoreEnv, self).__init__()

    def close(self):
        """
        Override in your subclass to perform any necessary cleanup.

        Environments will automatically close() themselves when
        garbage collected or when the program exits.
        """
        raise NotImplementedError

    def render(self, mode: str = 'human'):
        """
        Renders the game.

        The set of supported modes varies per game. (And some game do not support rendering at all.)

        # Arguments
            mode (str): The mode to render with. (default is 'human')
        """
        raise NotImplementedError

    def reset(self, *args):
        """
        Resets the state of the game and returns an initial observation.

        # Returns
            observation (abstract): The initial observation of the space. Initial reward is assumed to be 0.
        """
        raise NotImplementedError

    def seed(self, seed=None):
        """
        set the seed
        """
        raise NotImplementedError

    def step(self, action):
        """
        Run one timestep of the game's dynamics.

        Accepts an action and returns a tuple (observation, reward, done, info).

        # Arguments
            action (abstract): An action provided by the game.

        # Returns
            observation (abstract): Agent's observation of the current game.
            reward (float) : Amount of reward returned after previous action.
            done (boolean): Whether the episode has ended, in which case further step() calls will return undefined results.
            info (dict): Contains auxiliary diagnostic information (helpful for debugging, and sometimes learning).
        """
        raise NotImplementedError

    def goal(self):
        raise NotImplementedError



from __future__ import print_function

import datetime

from joatmon.assistant.task import BaseTask


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "clock",
            "description": "a function for user to learn current time and date",
            "parameters": {
                "type": "object",
                "properties": {
                },
                "required": []
            }
        }

    def run(self):
        self.api.output(datetime.datetime.now().isoformat())

        if not self.stop_event.is_set():
            self.stop_event.set()



import numpy as np

from joatmon.nn import functional as f
from joatmon.nn.core import (
    Module,
    Parameter,
    Tensor
)

__all__ = ['Conv']


class Conv(Module):
    def __init__(self, in_features, out_features, kernel_size, stride=1, padding=0):
        super(Conv, self).__init__()

        self._in_features = in_features
        self._out_features = out_features
        self._kernel_size = kernel_size
        self._stride = stride
        self._padding = padding

        self.weight = Parameter(Tensor.from_array(np.ones((out_features, in_features, *kernel_size))))
        self.bias = Parameter(Tensor.from_array(np.ones((out_features,))))

    def forward(self, inp):
        return f.conv(
            inp=inp,
            weight=self.weight,
            bias=self.bias,
            stride=self._stride,
            padding=self._padding,
        )



import dataclasses
import enum
import importlib.util
import json
import os
import threading

from transitions import Machine

from joatmon.utility import (
    first,
    JSONEncoder
)


class BaseService:
    def __init__(self, api, **kwargs):
        self.api = api
        self.kwargs = kwargs
        self.event = threading.Event()
        self.thread = threading.Thread(target=self.run)

        states = ['none', 'started', 'stopped', 'running', 'exception', 'starting', 'stopping']
        self.machine = Machine(model=self, states=states, initial='none')

    @staticmethod
    def help():
        return {}

    def run(self):
        raise NotImplementedError

    def running(self):
        return not self.event.is_set()

    def start(self):
        self.thread.start()

    def stop(self):
        self.event.set()


class ServiceState(enum.Enum):
    running = enum.auto()
    finished = enum.auto()
    stopped = enum.auto()


# create from json and to json methods
@dataclasses.dataclass
class ServiceInfo:
    name: str
    state: ServiceState
    service: BaseService


def on_begin(name, *args, **kwargs):
    ...


def on_error(name, *args, **kwargs):
    # if the service has recovery option, run it
    # else end
    ...


def on_end(name, *args, **kwargs):
    ...


def create(api):
    name = api.input('name')
    priority = api.input('priority')
    mode = api.input('mode')
    script = api.input('script')
    kwargs = {}
    for k in get_class(script).params():
        kwargs[k] = api.input(k)

    # on error
    # on recovery
    create_args = {
        'name': name,
        'priority': priority,
        'mode': mode,
        'script': script,
        'status': True,
        'kwargs': kwargs
    }

    settings = json.loads(open('iva/iva.json', 'r').read())
    services = settings.get('services', [])

    services.append(create_args)

    settings['services'] = services
    open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))


def get_class(name):
    service = None

    settings = json.loads(open('iva/iva.json', 'r').read())
    for scripts in settings.get('scripts', []):
        if os.path.isabs(scripts):
            if os.path.exists(scripts) and os.path.exists(os.path.join(scripts, f'{name}.py')):
                spec = importlib.util.spec_from_file_location(scripts, os.path.join(scripts, f'{name}.py'))
                action_module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(action_module)
            else:
                continue
        else:
            try:
                _module = __import__(scripts, fromlist=[f'{name}'])
            except ModuleNotFoundError:
                continue

            action_module = getattr(_module, name, None)

        if action_module is None:
            continue

        service = getattr(action_module, 'Service', None)

    return service


def get(api, name):
    settings = json.loads(open('iva/iva.json', 'r').read())
    task_info = first(filter(lambda x: x['status'] and x['name'] == name, settings.get('services', [])))

    if task_info is None:
        task_info = {'script': name, 'kwargs': {}}

    script = task_info['script']

    service = get_class(script)

    if service is None:
        api.output('service is not found')
        return None

    kwargs = {**task_info['kwargs'], 'parent_os_path': api.parent_os_path, 'os_path': api.os_path}

    service = service(api, **kwargs)
    return service



import json

from joatmon.assistant.task import BaseTask
from joatmon.utility import JSONEncoder


class Task(BaseTask):
    def __init__(self, api, **kwargs):
        super(Task, self).__init__(api, **kwargs)

    @staticmethod
    def help():
        return {
            "name": "config",
            "description": "a function for user to configure the iva",
            "parameters": {
                "type": "object",
                "properties": {
                    "action": {
                        "type": "string",
                        "enum": ["create", "update", "delete"]
                    },
                    "name": {
                        "type": "string",
                        "description": "name of the configuration"
                    },
                    "value": {
                        "type": "string",
                        "description": "value of the configuration"
                    }
                },
                "required": ["action", "name"]
            }
        }

    def run(self):
        action = self.kwargs.get('action', '') or self.api.input('what do you want to configure')
        assert action in ('create', 'update', 'delete')
        name = self.kwargs.get('name', '') or self.api.input('what is the name')
        value = self.kwargs.get('value', '') or self.api.input('what is the value')

        cfg = {
            'action': action,
            'name': name,
            'value': value
        }

        def set_config(_parent, _name, _value):
            if _name == '':
                return

            names = _name.split('.')
            if names[0] not in _parent:
                if len(names) == 1:
                    if _value is not None:
                        _parent[names[0]] = _value
                    else:
                        del _parent[names[0]]
                else:
                    _parent[names[0]] = {}
            else:
                if len(names) == 1:
                    if _value is not None:
                        _parent[names[0]] = _value
                    else:
                        del _parent[names[0]]
            set_config(_parent[names[0]], '.'.join(names[1:]), _value)

        settings = json.loads(open('iva/iva.json', 'r').read())
        config = settings.get('config', {})

        if cfg['action'] == 'create':
            set_config(config, cfg['name'], cfg['value'])
            settings['config'] = config
            open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))
        elif cfg['action'] == 'update':
            set_config(config, cfg['name'], cfg['value'])
            settings['config'] = config
            open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))
        elif cfg['action'] == 'delete':
            set_config(config, cfg['name'], None)
            settings['config'] = config
            open('iva/iva.json', 'w').write(json.dumps(settings, indent=4, cls=JSONEncoder))
        else:
            raise ValueError(f'arguments are not recognized')

        if not self.stop_event.is_set():
            self.stop_event.set()
